
import sys
import time
import os.path
import copy

from PyQt5.QtWidgets import (QApplication, QMainWindow,
        QDialog, QLabel, QTextEdit, QPushButton, QVBoxLayout,
        QSystemTrayIcon, QMenu, QAction, QActionGroup, QToolButton,
        QTreeView, QHeaderView, QInputDialog, QToolTip)
from PyQt5.QtCore import (QCoreApplication, QUrl, QBuffer, QIODevice, Qt,
    QTimer, QFile)
from PyQt5.QtCore import pyqtSignal, QUrl, QObject, QDateTime
from PyQt5 import QtWebEngineWidgets, QtWebEngine, QtWebEngineCore
from PyQt5.QtGui import (QClipboard, QPixmap, QIcon, QKeySequence,
        QStandardItemModel, QStandardItem)

from galacteek.core.asynclib import asyncify
from galacteek.ui import mediaplayer
from galacteek.ipfs.wrappers import *
from galacteek.ipfs.ipfsops import *
from galacteek.ipfs.cidhelpers import cidValid
from . import ui_galacteek, ui_ipfsinfos, ui_peersmgr
from . import (browser, files, keys, settings, hashmarks,
        textedit, ipfsview, dag, ipfssearchview, peers, eventlog)
from .helpers import *
from .modelhelpers import *
from .widgets import GalacteekTab
from .dialogs import *
from ..appsettings import *
from .i18n import *

def iHashmarks():
    return QCoreApplication.translate('GalacteekWindow', 'Hashmarks')
def iFileManager():
    return QCoreApplication.translate('GalacteekWindow', 'File Manager')
def iKeys():
    return QCoreApplication.translate('GalacteekWindow', 'IPFS Keys')

def iPinningStatus():
    return QCoreApplication.translate('GalacteekWindow', 'Pinning status')

def iDagViewer():
    return QCoreApplication.translate('GalacteekWindow', 'DAG viewer')

def iEventLog():
    return QCoreApplication.translate('GalacteekWindow', 'Event Log')

def iPeers():
    return QCoreApplication.translate('GalacteekWindow', 'Peers')

def iIpfsSearch(text):
    return QCoreApplication.translate('GalacteekWindow',
            'Search: {0}').format(text)

def iFromClipboard(path):
    return QCoreApplication.translate('GalacteekWindow',
        'Clipboard: browse IPFS path: {0}').format(path)

def iClipboardEmpty():
    return QCoreApplication.translate('GalacteekWindow',
        'No valid IPFS CID/path in the clipboard')

def iClipboardClearHistory():
    return QCoreApplication.translate('GalacteekWindow',
        'Clear clipboard history')

def iClipLoaderExplore(path):
    return QCoreApplication.translate('GalacteekWindow',
        'Explore IPFS path: {0}').format(path)

def iClipLoaderDagView(path):
    return QCoreApplication.translate('GalacteekWindow',
        'DAG view: {0}').format(path)

def iClipboardHistory():
    return QCoreApplication.translate('GalacteekWindow', 'Clipboard history')

def iNewProfile():
    return QCoreApplication.translate('GalacteekWindow', 'New Profile')

def iSwitchedProfile():
    return QCoreApplication.translate('GalacteekWindow',
            'Successfully switched profile')

def iClipLoaderBrowse(path):
    return QCoreApplication.translate('GalacteekWindow',
        'Browse IPFS path: {0}').format(path)

def iPinningItemStatus(pinPath, pinProgress):
    return QCoreApplication.translate('GalacteekWindow',
        '\nPath: {0}, nodes processed: {1}').format(pinPath, pinProgress)

def iAbout():
    from galacteek import __version__
    return QCoreApplication.translate('GalacteekWindow', '''
        <p><b>Galacteek</b> is a simple IPFS browser and content publisher
        </p>
        <p>Author: David Ferlier</p>
        <p>Galacteek version {0}</p>''').format(__version__)

class InfosDialog(QDialog):
    def __init__(self, app, parent=None):
        super().__init__(parent)

        self.ui = ui_ipfsinfos.Ui_IPFSInfosDialog()
        self.ui.setupUi(self)
        self.ui.okButton.clicked.connect(lambda: self.done(1))

    @ipfsOp
    async def loadInfos(self, ipfsop):
        try:
            repoStat = await ipfsop.client.repo.stat()
            idInfo   = await ipfsop.client.core.id()
        except:
            for label in [ self.ui.repoObjCount,
                    self.ui.repoVersion,
                    self.ui.repoSize,
                    self.ui.repoMaxStorage,
                    self.ui.nodeId,
                    self.ui.agentVersion,
                    self.ui.protocolVersion ]:
                label.setText(iUnknown())
        else:
            self.ui.repoObjCount.setText(str(repoStat.get(
                'NumObjects', iUnknown())))
            self.ui.repoVersion.setText(str(repoStat.get(
                'Version', iUnknown())))
            self.ui.repoSize.setText(sizeFormat(repoStat.get(
                'RepoSize', 0)))
            self.ui.repoMaxStorage.setText(sizeFormat(repoStat.get(
                'StorageMax', 0)))

            self.ui.nodeId.setText(idInfo.get('ID', iUnknown()))
            self.ui.agentVersion.setText(idInfo.get('AgentVersion', iUnknown()))
            self.ui.protocolVersion.setText(idInfo.get('ProtocolVersion',
                iUnknown()))

class PinStatusDetails(GalacteekTab):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)

        self.tree = QTreeView(self)
        self.vLayout = QVBoxLayout(self)
        self.vLayout.addWidget(self.tree)

        self.app.ipfsCtx.pinItemStatusChanged.connect(self.onPinStatusChanged)
        self.app.ipfsCtx.pinFinished.connect(self.onPinFinished)

        self.model = QStandardItemModel()
        self.model.setHorizontalHeaderLabels(
            ['Path', 'Nodes processed'])

        self.tree.setModel(self.model)
        self.tree.header().setSectionResizeMode(0,
                QHeaderView.ResizeToContents)

    def findPinItems(self, path):
        ret = modelSearch(self.model,
                search=path, columns=[0])
        if len(ret) == 0:
            return None, None
        itemP = self.model.itemFromIndex(ret.pop())
        idxS = self.model.index(itemP.row(), 1, itemP.index().parent())
        itemS = self.model.itemFromIndex(idxS)
        return itemP, itemS

    def onPinFinished(self, path):
        ePin, ePinS = self.findPinItems(path)
        if ePinS:
            ePinS.setText(iFinished())

    def onPinStatusChanged(self, path, status):
        nodesProcessed = status.get('Progress', None)
        ePin, ePinS = self.findPinItems(path)

        if not ePin:
            itemP = UneditableItem(path)
            itemS = UneditableItem(str(nodesProcessed) or iUnknown())
            self.model.invisibleRootItem().appendRow(
                    [itemP, itemS])
        else:
            if nodesProcessed:
                ePinS.setText(str(nodesProcessed))

class MainWindow(QMainWindow):
    def __init__(self, app):
        super(MainWindow, self).__init__()

        self.showMaximized()
        self._app = app
        self._allTabs = []

        self.ui = ui_galacteek.Ui_GalacteekWindow()
        self.ui.setupUi(self)

        self.tabnFManager = iFileManager()
        self.tabnKeys = iKeys()
        self.tabnPinning = iPinningStatus()
        self.tabnMediaPlayer = iMediaPlayer()

        self.ui.actionQuit.triggered.connect(self.quit)

        self.ui.actionCloseAllTabs.triggered.connect(
                self.onCloseAllTabs)
        self.ui.actionAboutGalacteek.triggered.connect(
                self.onAboutGalacteek)
        self.ui.actionDonate.triggered.connect(
                self.onHelpDonate)
        self.ui.actionSettings.triggered.connect(
                self.onSettings)
        self.ui.actionEvent_log.triggered.connect(
                self.onOpenEventLog)

        self.menuManual = QMenu(iManual())
        self.ui.menuAbout.addMenu(self.menuManual)

        self.ui.myFilesButton.clicked.connect(self.onFileManagerClicked)
        self.ui.myFilesButton.setShortcut(QKeySequence('Ctrl+f'))
        self.ui.manageKeysButton.clicked.connect(self.onIpfsKeysClicked)
        self.ui.openBrowserTabButton.clicked.connect(self.onOpenBrowserTabClicked)
        self.ui.hashmarksButton.clicked.connect(self.addHashmarksTab)
        self.ui.hashmarksButton.setShortcut(QKeySequence('Ctrl+m'))
        self.ui.writeNewDocumentButton.clicked.connect(self.onWriteNewDocumentClicked)
        self.ui.mediaPlayerButton.clicked.connect(self.onOpenMediaPlayer)
        self.ui.ipfsSearchButton.clicked.connect(self.onIpfsSearch)
        self.ui.ipfsSearch.returnPressed.connect(self.onIpfsSearch)
        self.ui.peersMgrButton.clicked.connect(self.onPeersMgrClicked)

        self.multiLoaderMenu = QMenu()
        self.multiLoaderHMenu = QMenu(iClipboardHistory())
        self.multiLoaderHMenu.triggered.connect(self.onClipboardHistoryMenu)
        self.multiLoadHashAction = QAction(getIconIpfsIce(),
                iClipboardEmpty(), self,
                shortcut=QKeySequence('Ctrl+o'),
                triggered=self.onLoadFromClipboard)
        self.multiExploreHashAction = QAction(getIconIpfsIce(),
                iClipboardEmpty(), self,
                shortcut=QKeySequence('Ctrl+e'),
                triggered=self.onExploreFromClipboard)
        self.multiDagViewAction = QAction(getIconIpfsIce(),
                iClipboardEmpty(), self,
                shortcut=QKeySequence('Ctrl+g'),
                triggered=self.onDagViewFromClipboard)

        self.multiExploreHashAction.setEnabled(False)
        self.multiLoadHashAction.setEnabled(False)
        self.multiDagViewAction.setEnabled(False)
        self.multiLoaderMenu.addAction(self.multiLoadHashAction)
        self.multiLoaderMenu.addAction(self.multiExploreHashAction)
        self.multiLoaderMenu.addAction(self.multiDagViewAction)
        self.multiLoaderMenu.addMenu(self.multiLoaderHMenu)

        self.ui.clipboardMultiLoader.clicked.connect(self.onLoadFromClipboard)
        self.ui.clipboardMultiLoader.setMenu(self.multiLoaderMenu)
        self.ui.clipboardMultiLoader.setToolTip(iClipboardEmpty())
        self.ui.clipboardMultiLoader.setPopupMode(QToolButton.MenuButtonPopup)

        # Global pin-all button
        self.ui.pinAllGlobalButton.setCheckable(True)
        self.ui.pinAllGlobalButton.setAutoRaise(True)
        self.pinAllGlobalChecked = False
        self.ui.pinAllGlobalButton.toggled.connect(self.onToggledPinAllGlobal)

        self.ui.tabWidget.setTabsClosable(True)
        self.ui.tabWidget.tabCloseRequested.connect(self.onTabCloseRequest)
        self.ui.tabWidget.setElideMode(Qt.ElideMiddle)
        self.ui.tabWidget.setUsesScrollButtons(True)

        self.webProfile = QtWebEngineWidgets.QWebEngineProfile.defaultProfile()

        self.ui.menuUser_Profile.addSeparator()
        self.ui.menuUser_Profile.triggered.connect(self.onUserProfile)
        self.profilesActionGroup = QActionGroup(self)

        # Status bar setup
        self.ui.pinningStatusButton = QPushButton()
        self.ui.pinningStatusButton.setToolTip(iNoStatus())
        self.ui.pinningStatusButton.setIcon(getIcon('pin-black.png'))
        self.ui.pinningStatusButton.clicked.connect(self.onPinningStatusDetails)
        self.ui.pubsubStatusButton = QPushButton()
        self.ui.pubsubStatusButton.setIcon(getIcon('network-offline.png'))
        self.ui.ipfsInfosButton = QPushButton()
        self.ui.ipfsInfosButton.setIcon(getIcon('information.png'))
        self.ui.ipfsInfosButton.setToolTip(iIpfsInfos())
        self.ui.ipfsInfosButton.clicked.connect(self.onIpfsInfos)

        self.profileMenu = QMenu()

        self.profileMenu.addAction('View Homepage',
                self.onProfileViewHomepage)
        self.profileMenu.addSeparator()
        self.profileMenu.addAction('Post Message',
                self.onProfilePostMessage)
        self.ui.profileEditButton.setEnabled(False)
        self.ui.profileEditButton.clicked.connect(self.onProfileEditDialog)
        self.ui.profileEditButton.setPopupMode(QToolButton.MenuButtonPopup)
        self.ui.profileEditButton.setMenu(self.profileMenu)

        self.ui.statusbar.addPermanentWidget(self.ui.ipfsInfosButton)
        self.ui.statusbar.addPermanentWidget(self.ui.pinningStatusButton)
        self.ui.statusbar.addPermanentWidget(self.ui.pubsubStatusButton)
        self.ui.statusbar.setStyleSheet('background-color: #4a9ea1')

        # Connection status timer
        self.timerStatus = QTimer(self)
        self.timerStatus.timeout.connect(self.onMainTimerStatus)
        self.timerStatus.start(6000)

        self.enableButtons(False)

        # Connect the IPFS context signals
        self.app.ipfsCtx.ipfsConnectionReady.connect(self.onConnReady)
        self.app.ipfsCtx.ipfsRepositoryReady.connect(self.onRepoReady)
        self.app.ipfsCtx.pubsub.psMessageRx.connect(self.onPubsubRx)
        self.app.ipfsCtx.pubsub.psMessageTx.connect(self.onPubsubTx)
        self.app.ipfsCtx.profilesAvailable.connect(self.onProfilesList)
        self.app.ipfsCtx.profileChanged.connect(self.onProfileChanged)
        self.app.ipfsCtx.pinItemStatusChanged.connect(self.onPinStatusChanged)
        self.app.ipfsCtx.pinItemsCount.connect(self.onPinItemsCount)

        # Application signals
        self.app.clipTracker.clipboardHasIpfs.connect(self.onClipboardIpfs)
        self.app.clipTracker.clipboardHistoryChanged.connect(self.onClipboardHistory)
        self.app.manualAvailable.connect(self.onManualAvailable)

        self.ui.tabWidget.removeTab(0)

    @property
    def app(self):
        return self._app

    @property
    def allTabs(self):
        return self._allTabs

    def onOpenEventLog(self):
        self.addEventLogTab(current=True)

    def onProfileEditDialog(self):
        runDialog(ProfileEditDialog, self.app.ipfsCtx.currentProfile,
                title='Profile Edit dialog')

    @asyncify
    async def onProfileChanged(self, pName, profile):
        self.ui.profileEditButton.setEnabled(False)
        await profile.userInfo.loaded
        self.ui.profileEditButton.setEnabled(True)

        for action in self.profilesActionGroup.actions():
            if action.data() == pName:
                action.setChecked(True)
                # Refresh the file manager
                filesM = self.findTabFileManager()
                if filesM:
                    filesM.setupModel()
                    filesM.pathSelectorDefault()

    def onUserProfile(self, action):
        if action is self.ui.actionNew_Profile:
            inText = QInputDialog.getText(self, iNewProfile(),
                    iNewProfile())
            profile, create = inText
            if create is True and profile:
                self.app.task(self.app.ipfsCtx.profileNew, profile,
                        emitavail=True)
        else:
            pName = action.text()
            if action.isChecked() and \
                    self.app.ipfsCtx.currentProfile.name != pName:
                rCode = self.app.ipfsCtx.profileChange(pName)

    def onProfileCreateGPGKey(self):
        pass

    def onProfilePostMessage(self):
        runDialog(ProfilePostMessageDialog, self.app.ipfsCtx.currentProfile)

    def onProfileViewHomepage(self):
        #self.addBrowserTab().browseIpnsHash(self.app.ipfsCtx.currentProfile.keyRootId)
        self.addBrowserTab().browseFsPath(os.path.join(
            joinIpns(self.app.ipfsCtx.currentProfile.keyRootId), 'index.html'))

    def onProfilesList(self, pList):
        currentList = [action.data() for action in \
                self.profilesActionGroup.actions()]

        currentProfile = self.app.ipfsCtx.currentProfile

        for pName in pList:
            if pName in currentList:
                continue

            action = QAction(self.profilesActionGroup,
                    checkable=True, text=pName)
            action.setData(pName)

        for action in self.profilesActionGroup.actions():
            self.ui.menuUser_Profile.addAction(action)

    def onRepoReady(self):
        self.enableButtons()

    def onIpfsInfos(self):
        dlg = InfosDialog(self.app)
        dlg.setWindowTitle(iIpfsInfos())
        self.app.task(dlg.loadInfos)
        dlg.exec_()

    def onConnReady(self):
        if self.app.debugEnabled is True:
            self.addEventLogTab()

        # Peers manager
        self.showPeersMgr()

    def onPubsubRx(self):
        now = QDateTime.currentDateTime()
        self.ui.pubsubStatusButton.setIcon(getIcon('network-transmit.png'))
        self.ui.pubsubStatusButton.setToolTip(
                'Pubsub: last message received {}'.format(now.toString()))

    def onPubsubTx(self):
        pass

    def onPinningStatusDetails(self):
        name = self.tabnPinning
        ft = self.findTabWithName(name)
        if ft:
            return self.ui.tabWidget.setCurrentWidget(ft)
        detailsTab = PinStatusDetails(self)
        self.registerTab(detailsTab, name)

    def updatePinningStatus(self):
        iconLoading = getIcon('pin-blue-loading.png')
        iconNormal = getIcon('pin-black.png')

        status = copy.copy(self.app.ipfsCtx.pinner.status)
        statusMsg = iItemsInPinningQueue(len(status))

        for pinPath, pinStatus in status.items():
            pinProgress = 'unknown'
            if pinStatus:
                pinProgress = pinStatus.get('Progress', 'unknown')

            statusMsg += iPinningItemStatus(pinPath, pinProgress)

        self.ui.pinningStatusButton.setToolTip(statusMsg)
        self.ui.pinningStatusButton.setStatusTip(statusMsg)

        del status

    def onPinItemsCount(self, count):
        iconLoading = getIcon('pin-blue-loading.png')
        iconNormal = getIcon('pin-black.png')

        if count > 0:
            self.ui.pinningStatusButton.setIcon(iconLoading)
        else:
            self.ui.pinningStatusButton.setIcon(iconNormal)

        self.updatePinningStatus()

    def onPinFinished(self, path):
        self.app.systemTrayMessage('PIN', iPinSuccess(path))

    def onPinStatusChanged(self, path, status):
        self.updatePinningStatus()

    def onManualAvailable(self, lang, entry):
        self.menuManual.addAction(lang, lambda:
            self.onOpenManual(lang, entry))

    def onOpenManual(self, lang, docEntry):
        self.addBrowserTab().browseIpfsHash(docEntry['Hash'])

    def enableButtons(self, flag=True):
        for btn in [ self.ui.myFilesButton,
                self.ui.manageKeysButton,
                self.ui.openBrowserTabButton,
                self.ui.hashmarksButton,
                self.ui.mediaPlayerButton,
                self.ui.writeNewDocumentButton,
                self.ui.peersMgrButton,
                self.ui.clipboardMultiLoader ]:
            btn.setEnabled(flag)

    def statusMessage(self, msg):
        self.ui.statusbar.showMessage(msg)

    def registerTab(self, tab, name, icon=None, current=False, add=True,
            tooltip=None):
        idx = None

        if add is True:
            if icon:
                idx = self.ui.tabWidget.addTab(tab, icon, name)
            else:
                idx = self.ui.tabWidget.addTab(tab, name)

        self._allTabs.append(tab)

        if current is True:
            self.ui.tabWidget.setCurrentWidget(tab)

        if tooltip and idx:
            self.ui.tabWidget.setTabToolTip(idx, tooltip)

    def findTabFileManager(self):
        return self.findTabWithName(self.tabnFManager)

    def findTabIndex(self, w):
        return self.ui.tabWidget.indexOf(w)

    def findTabWithName(self, name):
        for idx in range(0, self.ui.tabWidget.count()):
            tName = self.ui.tabWidget.tabText(idx)

            if tName == name:
                return self.ui.tabWidget.widget(idx)

    def removeTabFromWidget(self, w):
        idx = self.ui.tabWidget.indexOf(w)
        if idx:
            self.ui.tabWidget.removeTab(idx)

    def onSettings(self):
        runDialog(settings.SettingsDialog, self.app)

    def onCloseAllTabs(self):
        self.ui.tabWidget.clear()

    def onToggledPinAllGlobal(self, checked):
        self.pinAllGlobalChecked = checked

    def onClipboardHistoryMenu(self, action):
        aData = action.data()

        if aData:
            hItem = aData['item']
            self.addBrowserTab().browseFsPath(hItem['path'])

    def onClipboardHistory(self, history):
        # Called when the clipboard history has changed
        self.multiLoaderHMenu.clear()
        hItems = history.items()

        def onHistoryClear():
            self.app.clipTracker.clearHistory()
            self.onClipboardIpfs(False, None, None)

        self.multiLoaderHMenu.addSeparator()
        self.multiLoaderHMenu.addAction(iClipboardClearHistory(),
                lambda: onHistoryClear())

        for hTs, hItem in hItems:
            action = QAction('{0} ({1})'.format(hItem['path'],
                        hItem['date'].toString()), self)
            action.setIcon(getIconIpfsIce())
            action.setData({
                'item': hItem
            })

            self.multiLoaderHMenu.addAction(action)

    def onClipboardIpfs(self, valid, cid, path):
        self.multiExploreHashAction.setEnabled(valid)
        self.multiLoadHashAction.setEnabled(valid)
        self.multiDagViewAction.setEnabled(valid)
        if valid:
            self.multiExploreHashAction.setText(iClipLoaderExplore(path))
            self.multiLoadHashAction.setText(iClipLoaderBrowse(path))
            self.multiDagViewAction.setText(iClipLoaderDagView(path))
            self.ui.clipboardMultiLoader.setToolTip(iFromClipboard(path))
        else:
            self.multiExploreHashAction.setText(iClipboardEmpty())
            self.multiLoadHashAction.setText(iClipboardEmpty())
            self.multiDagViewAction.setText(iClipboardEmpty())
            self.ui.clipboardMultiLoader.setToolTip(iClipboardEmpty())

    def onLoadFromClipboard(self):
        current = self.app.clipTracker.getCurrent()
        if current:
            self.addBrowserTab().browseFsPath(current['path'])
        else:
            messageBox(iClipboardEmpty())

    def onExploreFromClipboard(self):
        current = self.app.clipTracker.getCurrent()
        if current:
            self.app.task(self.exploreClipboardPath, current['path'])
        else:
            messageBox(iClipboardEmpty())

    def onDagViewFromClipboard(self):
        current = self.app.clipTracker.getCurrent()
        if current:
            view = dag.DAGViewer(current['path'], self)
            self.registerTab(view, iDagViewer(), current=True)
        else:
            messageBox(iClipboardEmpty())

    @ipfsStatOp
    async def exploreClipboardPath(self, ipfsop, path, stat):
        if stat:
            self.exploreHash(stat['Hash'])

    def onAboutGalacteek(self):
        from galacteek import __version__
        QMessageBox.about(self, 'About Galacteek', iAbout())

    @asyncify
    async def onMainTimerStatus(self):
        @ipfsOpFn
        async def connectionInfo(oper):
            try:
                info = await oper.client.core.id()
            except:
                return self.statusMessage(iErrNoCx())

            nodeId = info.get('ID', iUnknown())
            nodeAgent = info.get('AgentVersion', iUnknownAgent())

            # Get IPFS peers list
            peers = await oper.peersList()
            if not peers:
                return self.statusMessage(iCxButNoPeers(
                    nodeId, nodeAgent))

            message = iConnectStatus(nodeId, nodeAgent, len(peers))
            self.statusMessage(message)

        await connectionInfo()

    def keyPressEvent(self, event):
        # Ultimately this will be moved to configurable shortcuts
        modifiers = event.modifiers()

        if modifiers & Qt.ControlModifier:
            if event.key() == Qt.Key_T:
                self.addBrowserTab()
            if event.key() == Qt.Key_W:
                idx = self.ui.tabWidget.currentIndex()
                self.onTabCloseRequest(idx)

        super(MainWindow, self).keyPressEvent(event)

    def exploreHash(self, hashV):
        tabName = '... {0}'.format(hashV[2*int(len(hashV)/3):])
        tooltip = 'Hash explorer: {0}'.format(hashV)
        view = ipfsview.IPFSHashExplorerToolBox(self, hashV)
        tab = self.registerTab(view, tabName, current=True,
                icon=getIcon('hash.png'), tooltip=tooltip)

    def addMediaPlayerTab(self):
        name = self.tabnMediaPlayer
        ft = self.findTabWithName(name)
        if ft:
            return ft
        tab = mediaplayer.MediaPlayerTab(self)
        self.registerTab(tab, name, icon=getIcon('multimedia.png'),
                current=True)
        return tab

    def mediaPlayerQueue(self, path, mediaName=None):
        tab = self.addMediaPlayerTab()
        tab.queueFromPath(path, mediaName=mediaName)

    def mediaPlayerPlay(self, path, mediaName=None):
        tab = self.addMediaPlayerTab()
        tab.playFromPath(path, mediaName=mediaName)

    def addHashmarksTab(self):
        name = iHashmarks()
        ft = self.findTabWithName(name)
        if ft:
            return self.ui.tabWidget.setCurrentWidget(ft)

        tab = hashmarks.HashmarksTab(self)
        self.registerTab(tab, name, current=True)

    def onTabCloseRequest(self, idx):
        tabName = self.ui.tabWidget.tabText(idx)
        tab = self.ui.tabWidget.widget(idx)

        if not tab in self.allTabs:
            return False

        if tab.onClose() is True:
            self.ui.tabWidget.removeTab(idx)
            del tab

    def onIpfsSearch(self):
        text = self.ui.ipfsSearch.text()
        self.ui.ipfsSearch.clear()
        if len(text) > 0:
            view = ipfssearchview.IPFSSearchView(text, self)
            self.registerTab(view, iIpfsSearch(text), current=True,
                    icon=getIcon('ipfs-search.png'))

    def onOpenMediaPlayer(self):
        self.addMediaPlayerTab()

    def onOpenBrowserTabClicked(self):
        self.addBrowserTab()

    def onWriteNewDocumentClicked(self):
        w = textedit.AddDocumentWidget(self, parent=self.ui.tabWidget)
        self.registerTab(w, 'New document', current=True)

    def onPeersMgrClicked(self):
        self.showPeersMgr(current=True)

    def onFileManagerClicked(self):
        name = self.tabnFManager

        icon = getIcon('folder-open.png')
        ft = self.findTabWithName(name)
        if ft:
            ft.updateTree()
            return self.ui.tabWidget.setCurrentWidget(ft)

        filesTab = files.FilesTab(self, parent=self.ui.tabWidget)
        self.registerTab(filesTab, name, current=True, icon=icon)

        filesTab.updateTree()

    def onIpfsKeysClicked(self):
        name = self.tabnKeys
        ft = self.findTabWithName(name)
        if ft:
            return self.ui.tabWidget.setCurrentWidget(ft)

        keysTab = keys.KeysTab(self)
        self.registerTab(keysTab, name, current=True)

    def onHelpDonate(self):
        bcAddress = '3HSsNcwzkiWGu6wB18BC6D37JHExpxZvyS'
        runDialog(DonateDialog, bcAddress)

    def addBrowserTab(self, label='No page loaded'):
        icon = getIconIpfsIce()
        tab = browser.BrowserTab(self, parent=self.ui.tabWidget)
        self.ui.tabWidget.addTab(tab, icon, label)
        self.ui.tabWidget.setCurrentWidget(tab)

        mgr = self.app.settingsMgr
        if mgr.isTrue(CFG_SECTION_BROWSER, CFG_KEY_GOTOHOME):
            tab.loadHomePage()

        self.allTabs.append(tab)
        return tab

    def addEventLogTab(self, current=False):
        self.registerTab(eventlog.EventLogWidget(self), iEventLog(),
                current=current)

    def showPeersMgr(self, current=False):
        # Peers mgr
        name = iPeers()

        ft = self.findTabWithName(name)
        if ft:
            return self.ui.tabWidget.setCurrentWidget(ft)

        self.peersTracker = peers.PeersTracker(self.app.ipfsCtx)
        pMgr = peers.PeersManager(self, self.peersTracker)
        self.registerTab(pMgr, name, current=current)

    def quit(self):
        # Qt and application exit
        self.app.onExit()

    def closeEvent(self, event):
        event.ignore()
        self.hide()
        self.app.systemTrayMessage('Galacteek', iMinimized())




import os.path

from PyQt5 import QtWebEngineWidgets, QtWebEngine, QtWebEngineCore
from PyQt5.QtWidgets import QMessageBox
from PyQt5.QtCore import QCoreApplication, QObject

from ..appsettings import *
from .i18n import *

def iFinishedDownload(filename):
    return QCoreApplication.translate('Galacteek',
        'Finished downloading file {0}').format(filename)

def iStartingDownload(filename):
    return QCoreApplication.translate('Galacteek',
        'Downloading file {0} ..').format(filename)

class DownloadsManager(QObject):
    def __init__(self, app):
        super(DownloadsManager, self).__init__()

        self.app = app
        self.webProfile = QtWebEngineWidgets.QWebEngineProfile.defaultProfile()
        self.webProfile.downloadRequested.connect(self.onDownloadRequested)

    def onDownloadRequested(self, downItem):
        if not downItem:
            return

        downloadsLoc = self.app.settingsMgr.eGet(S_DOWNLOADS_PATH)

        def progress(received, total):
            pass

        def finished(item):
            filename = item.path() or iUnknown()
            self.app.systemTrayMessage('Galacteek', iFinishedDownload(filename))

        name = os.path.basename(downItem.path())

        self.app.systemTrayMessage('Galacteek', iStartingDownload(name))

        downItem.setPath(os.path.join(downloadsLoc, name))
        downItem.finished.connect(lambda: finished(downItem))
        downItem.downloadProgress.connect(progress)
        downItem.accept()




# You need the pytest-qt plugin to run these tests

import pytest
import asyncio
import os.path

import tempfile
import time
import random

from PyQt5.QtCore import Qt

from .cidtest import *

from galacteek.ui import dialogs
from galacteek.ui import mainui, ipfsview
from galacteek import application
from galacteek.appsettings import *
from galacteek.ipfs.wrappers import *
from galacteek.ipfs.ipfsops import *

class TestCIDDialogs:
    # Valid and invalid CID lists
    @pytest.mark.parametrize('validcid0', validcid0)
    @pytest.mark.parametrize('validcid1', validcid1)
    @pytest.mark.parametrize('invalidcid', invalidcids)
    def test_singlecidinput(self, qtbot, validcid0, validcid1, invalidcid):
        dlg = dialogs.IPFSCIDInputDialog()
        qtbot.addWidget(dlg)

        self.cidInput(qtbot, dlg, validcid0)
        assert dlg.validCid is True
        assert dlg.ui.validity.text() == 'Valid CID version 0'

        dlg.clearCID()
        self.cidInput(qtbot, dlg, validcid1)
        assert dlg.validCid is True
        assert dlg.ui.validity.text() == 'Valid CID version 1'

        dlg.clearCID()
        self.cidInput(qtbot, dlg, invalidcid)
        assert dlg.validCid is False

    @pytest.mark.parametrize('validcid0', validcid0)
    def test_mulcidinput(self, qtbot, validcid0):
        dlg = dialogs.IPFSMultipleCIDInputDialog()
        qtbot.addWidget(dlg)

        self.cidInput(qtbot, dlg, validcid0)
        qtbot.mouseClick(dlg.ui.addCidButton, Qt.LeftButton)

        cidList = dlg.getCIDs()
        assert len(cidList) == 1
        assert cidList.pop() == validcid0

    def cidInput(self, qtbot, dlg, cid):
        # Manually input the CID character by character
        # The whole unit test is a bit cpu-intensive since the dialog will
        # check CID's validity on every char input but hey ..
        for char in cid:
            qtbot.keyPress(dlg.ui.cid, char)

@pytest.fixture
def testfile1(tmpdir):
    filep = tmpdir.join('testfile1.txt')
    filep.write('POIEKJDOOOPIDMWOPIMPOWE()=ds129084bjcy')
    return filep

@pytest.fixture
def dirTestFiles(pytestconfig):
    return os.path.join(pytestconfig.rootdir,
        'test', 'testfiles')

@pytest.fixture
def testFilesDocsAsync(dirTestFiles):
    return os.path.join(dirTestFiles, 'docs-asyncio')

def makeAppNoDaemon():
    gApp = application.GalacteekApplication(profile='pytest-noipfsd', debug=True)
    loop = gApp.setupAsyncLoop()
    sManager = gApp.settingsMgr
    sManager.setFalse(CFG_SECTION_IPFSD, CFG_KEY_ENABLED)
    sManager.sync()
    gApp.updateIpfsClient()
    return gApp

def makeApp(profile='pytest-withipfsd'):
    r = random.Random()
    gApp = application.GalacteekApplication(profile=profile, debug=True)
    loop = gApp.setupAsyncLoop()

    sManager = gApp.settingsMgr
    sManager.setTrue(CFG_SECTION_IPFSD, CFG_KEY_ENABLED)
    sManager.setSetting(CFG_SECTION_IPFSD, CFG_KEY_APIPORT,
            r.randint(15500, 15580))
    sManager.setSetting(CFG_SECTION_IPFSD, CFG_KEY_SWARMPORT,
            r.randint(15600, 15680))
    sManager.setSetting(CFG_SECTION_IPFSD, CFG_KEY_HTTPGWPORT,
            r.randint(15700, 15780))
    sManager.sync()

    return gApp

@pytest.fixture
#@pytest.mark.asyncio
def gApp(qtbot, tmpdir):
    gApp = makeApp()
    sManager = gApp.settingsMgr
    sManager.setSetting(CFG_SECTION_BROWSER, CFG_KEY_DLPATH, str(tmpdir))
    sManager.sync()

    gApp.startIpfsDaemon()
    with qtbot.waitSignal(gApp.ipfsCtx.ipfsRepositoryReady, timeout=20000):
        print('Waiting for IPFS repository ...')

    yield gApp

    gApp.ipfsd.stop()
    gApp.stopIpfsServices()
    gApp.quit()

@pytest.fixture(scope='module')
def modApp():
    return makeApp(profile='pytest-modapp')

class TestApp:
    @pytest.mark.asyncio
    async def test_appmain(self, qtbot, gApp, tmpdir, testfile1):
        """
        Test the application, running the GUI and testing all components
        """
        loop = gApp.loop

        def leftClick(w):
            qtbot.mouseClick(w, Qt.LeftButton)

        with loop:
            mainW = gApp.mainWindow

            async def openTabs():
                leftClick(mainW.ui.myFilesButton)
                leftClick(mainW.ui.openBrowserTabButton)
                leftClick(mainW.ui.hashmarksButton)
                leftClick(mainW.ui.manageKeysButton)
                leftClick(mainW.ui.writeNewDocumentButton)

                # Activate global pinning
                leftClick(mainW.ui.pinAllGlobalButton)
                await asyncio.sleep(1)

            async def browse():
                # Browse some hashes retrieved from the file manager's entry
                # cache, opening a tab for every tab
                await asyncio.sleep(1)
                filesW = mainW.findTabFileManager()
                cacheDirs = filesW.model.entryCache.getByType(1)
                hashes = list(cacheDirs.keys())

                for ohash in hashes[0:16]:
                    t = gApp.mainWindow.addBrowserTab()
                    t.browseIpfsHash(ohash)
                await asyncio.sleep(5)

            loop.run_until_complete(openTabs())
            loop.run_until_complete(browse())

    @pytest.mark.asyncio
    async def test_ipfsview(self, qtbot, gApp, testFilesDocsAsync):
        loop = gApp.loop

        @ipfsOpFn
        async def addAndView(ipfsop):
            view = ipfsview.IPFSHashExplorerToolBox(gApp.mainWindow, None)
            gApp.mainWindow.registerTab(view, '')
            root = await ipfsop.addPath(testFilesDocsAsync)
            view.viewHash(root['Hash'], autoOpenFolders=True)
            await ipfsop.sleep(2)

        with loop:
            loop.run_until_complete(addAndView())

    @pytest.mark.parametrize('validcid0', validcid0)
    @pytest.mark.parametrize('validcid1', validcid1)
    @pytest.mark.parametrize('invalidcid', invalidcids)
    def test_clipboard(self, qtbot, modApp, validcid0, validcid1, invalidcid):
        def checkCorrect(valid, cid, path):
            item = modApp.clipTracker.getHistoryLatest()
            assert item['path'] == path
            return valid == True
        def checkIncorrect(valid, cid, path):
            return valid == False

        def waitCorrect(cb=checkCorrect, timeout=1000):
            return qtbot.waitSignal(modApp.clipTracker.clipboardHasIpfs,
                timeout=timeout, check_params_cb=cb)

        def waitIncorrect(cb=checkIncorrect, timeout=1000):
            return qtbot.waitSignal(modApp.clipTracker.clipboardHasIpfs,
                timeout=timeout, check_params_cb=cb)

        modApp.clipTracker.clearHistory()

        # Bare CIDv0
        with waitCorrect():
            modApp.setClipboardText(validcid0)

        # Multiline CIDs
        with waitCorrect():
            modApp.setClipboardText('\n'.join([
                validcid0,
                '++' * 30,
                validcid1
            ]))

        # Bare CIDv1
        with waitCorrect():
            modApp.setClipboardText(validcid1)

        modApp.clipTracker.clearHistory()

        # /ipfs/CID
        with waitCorrect():
            modApp.setClipboardText(joinIpfs(validcid0))

        # /ipfs/CID with leading and trailing spaces
        with waitCorrect():
            modApp.setClipboardText('      {}   '.format(joinIpfs(validcid0)))

        # http(s)://ipfs.io/ipfs/CID
        with waitCorrect():
            modApp.setClipboardText('https://ipfs.io{}'.format(joinIpfs(validcid0)))

        # http(s)://localhost:8080/ipfs/CID
        with waitCorrect():
            modApp.setClipboardText('http://localhost:8080{}'.format(joinIpfs(validcid0)))

        # http(s)://localhost:8080/ipns/CID
        with waitCorrect():
            modApp.setClipboardText('http://localhost:8080{}'.format(joinIpns(validcid1)))

        # multiple lines with /ipfs/CID, will only register the first match
        with waitCorrect():
            modApp.setClipboardText('\n'.join([
                joinIpfs(validcid0),
                joinIpfs(validcid1),
            ]))

        # fs:/ipfs/CID
        with waitCorrect():
            modApp.setClipboardText('fs:' + joinIpfs(validcid0))

        # ipfs:/ipfs/CID
        with waitCorrect():
            modApp.setClipboardText('ipfs:' + joinIpfs(validcid0))

        # /ipfs/CID/something
        with waitCorrect():
            path = os.path.join(joinIpfs(validcid0), 'some', 'garlic')
            modApp.setClipboardText(path)

        # /ipns/CID/something
        with waitCorrect():
            path = os.path.join(joinIpns(validcid0), 'share', 'the', 'wine')
            modApp.setClipboardText(path)

        # Invalid CID input
        with waitIncorrect():
            modApp.setClipboardText(invalidcid)

    def test_mediaplayer(self, qtbot, modApp):
        pass




import sys
import os.path
import json
import tempfile
import uuid

from async_generator import async_generator, yield_
from yarl import cached_property

from galacteek import log, ensure

import aioipfs
import asyncio

GFILES_ROOT_PATH = '/galacteek/'

def joinIpfs(path):
    return os.path.join('/ipfs/', path)

def stripIpfs(path):
    return path.strip('/ipfs/')

def joinIpns(path):
    return os.path.join('/ipns/', path)

def isDict(data):
    return isinstance(data, dict)

class OperatorError(Exception):
    pass

class IPFSLogWatcher(object):
    def __init__(self, operator):
        self.op = operator

    async def analyze(self):
        import pprint
        async for msg in self.op.client.log.tail():
            event = msg.get('event', None)
            time = msg.get('time', None)

            if not event:
                continue

            if event == 'handleAddProvider':
                # Handle add provider
                if self.op.ctx:
                    self.op.ctx.logAddProvider.emit(msg)

class IPFSOperator(object):
    """
    IPFS operator, for your daily operations!
    """

    def __init__(self, client, ctx=None, debug=False):
        self._id = uuid.uuid1()
        self._cache = {}
        self.client = client
        self.debugInfo = debug
        self.ctx = ctx

        self.filesChroot = None
        self._commands = None

        self.evReady = asyncio.Event()

    @property
    def uid(self):
        return self._id

    @property
    def logwatch(self):
        return IPFSLogWatcher(self)

    @property
    def availCommands(self):
        """ Cached property: available IPFS commands """
        return self._commands

    def debug(self, msg):
        log.debug('IPFSOp({0}): {1}'.format(self.uid, msg))

    async def sleep(self, t=0):
        await asyncio.sleep(t)

    async def waitReady(self, timeout=10):
        return await self.waitFor(self.evReady.wait(), timeout)

    async def waitFor(self, fncall, timeout):
        try:
            output = await asyncio.wait_for(fncall, timeout)
        except asyncio.TimeoutError as timeExc:
            self.debug('Timeout waiting for coroutine {0}'.format(fncall))
            return None
        except aioipfs.APIError as e:
            self.debug(e.message)
            return None
        else:
            return output

    async def getCommands(self):
        if self.availCommands is not None:
            return self.availCommands
        try:
            self._commands = await self.client.core.commands()
        except aioipfs.APIError as err:
            self.debug('Cannot find available IPFS commands')
            return None
        else:
            return self.availCommands

    async def filesDelete(self, path, name, recursive=False):
        try:
            ret = await self.client.files.rm(os.path.join(path, name),
                    recursive=recursive)
        except aioipfs.APIError as err:
            self.debug('Exception on removing {0} in {1}'.format(
                name, path))
            return False
        await self.client.files.flush(path)
        return True

    async def filesLookup(self, path, name):
        """
        Looks for a file (files API) with a given name in this path

        :param str path: the path to search
        :param str name: the entry name to look for
        :return: IPFS entry
        """
        listing = await self.filesList(path)
        if listing is None:
            return
        for entry in listing:
            if entry['Name'] == name:
                return entry

    async def filesLookupHash(self, path, hash):
        """
        Searches for a file (files API) with a given hash in this path

        :param str path: the path to search
        :param str hash: the multihash to look for
        :return: IPFS entry
        """
        listing = await self.filesList(path)
        for entry in listing:
            if entry['Hash'] == hash:
                return entry

    async def filesRm(self, path, recursive=False):
        try:
            await self.client.files.rm(path, recursive=recursive)
        except aioipfs.APIError as err:
            self.debug(err.message)
            return None

    async def filesMkdir(self, path, parents=True):
        try:
            await self.client.files.mkdir(path, parents=parents)
        except aioipfs.APIError as err:
            self.debug(err.message)
            return None

    async def filesMove(self, source, dest):
        try:
            return await self.client.files.mv(source, dest)
        except aioipfs.APIError as err:
            self.debug(err.message)
            return None

    async def filesList(self, path):
        try:
            listing = await self.client.files.ls(path, long=True)
        except aioipfs.APIError as err:
            self.debug(err.message)
            return None

        if not 'Entries' in listing or listing['Entries'] is None:
            return []

        return listing['Entries']

    async def filesStat(self, path):
        try:
            return await self.client.files.stat(path)
        except aioipfs.APIError as err:
            self.debug(err.message)
            return None

    async def filesWrite(self, path, data, create=False, truncate=False,
            offset=-1, count=-1):
        try:
            resp = await self.client.files.write(path, data,
                    create=create, truncate=truncate,
                    offset=offset, count=count)
        except aioipfs.APIError as err:
            self.debug('filesWrite error {}'.format(err.message))
            return None
        else:
            self.debug('filesWrite success {}'.format(resp))
            return resp

    async def filesWriteJsonObject(self, path, obj, create=True,
            truncate=True):
        try:
            serialized = json.dumps(obj).encode()
            resp = await self.filesWrite(path, serialized,
                    create=create, truncate=truncate)
        except aioipfs.APIError as err:
            self.debug('filesWriteJson error {}'.format(err.message))
            return None
        else:
            return resp

    async def filesReadJsonObject(self, path):
        try:
            resp = await self.client.files.read(path)
        except aioipfs.APIError as err:
            self.debug('filesReadJson error {}'.format(err.message))
            return None
        except Exception as err:
            self.debug('filesReadJson unknown error {}'.format(str(err)))
            return None
        else:
            return json.loads(resp.decode())

    async def chroot(self, path):
        self.filesChroot = path

    async def vMkdir(self, path):
        if self.filesChroot:
            return await self.filesMkdir(os.path.join(self.filesChroot,
                path))

    async def vFilesList(self, path):
        if self.filesChroot:
            return await self.filesList(os.path.join(self.filesChroot,
                path))
        else:
            raise OperatorError('No chroot provided')

    async def filesCp(self, srcHash, dest):
        try:
            resp = await self.client.files.cp(joinIpfs(srcHash),
                    dest)
        except aioipfs.APIError as err:
            return False

        return True

    async def filesLink(self, entry, dest, flush=True, name=None):
        """ Given an entry (as returned by /files/ls), make a link
            in dest """

        try:
            resp = await self.client.files.cp(joinIpfs(entry['Hash']),
                os.path.join(dest, name if name else entry['Name']))
        except aioipfs.APIError as err:
            self.debug('Exception on copying entry {0} to {1}'.format(
                entry, dest))
            return False

        if flush:
            await self.client.files.flush(dest)
        return True

    async def filesLinkFp(self, entry, dest):
        try:
            resp = await self.client.files.cp(joinIpfs(entry['Hash']),
                    dest)
        except aioipfs.APIError as err:
            self.debug('Exception on copying entry {0} to {1}'.format(
                entry, dest))
            return False

        return True

    async def peersList(self):
        try:
            peers = await self.client.swarm.peers()
        except aioipfs.APIError as err:
            return None

        if isDict(peers) and 'Peers' in peers:
            return peers['Peers']

    async def nodeId(self):
        info = await self.client.core.id()
        if isDict(info):
            return info.get('ID', 'Unknown')

    async def keysNames(self):
        keys = await self.keys()
        return [key['Name'] for key in keys]

    async def keyGen(self, keyName, type='rsa', keySize=2048):
        return await self.waitFor(self.client.key.gen(keyName,
                type=type, size=keySize), 30)

    async def keys(self):
        kList = await self.client.key.list(long=True)
        if isDict(kList):
            return kList.get('Keys', [])

    async def keyFind(self, name):
        keys = await self.keys()
        for key in keys:
            if key['Name'] == name:
                return key

    async def keysRemove(self, name):
        try:
            await self.client.key.rm(name)
        except aioipfs.APIError as err:
            self.debug('Exception on removing key {0}: {1}'.format(
                name, err.message))
            return False
        return True

    async def publish(self, path, key='self', timeout=90):
        return await self.waitFor(self.client.name.publish(path, key=key),
                timeout)

    async def resolve(self, path, timeout=20):
        try:
            resolved = await asyncio.wait_for(self.client.name.resolve(path),
                    timeout)
        except asyncio.TimeoutError as timeExc:
            return None
        except aioipfs.APIError as e:
            self.debug('resolve error: {}'.format(e.message))
            return None
        else:
            return resolved

    async def purge(self, hashRef, rungc=False):
        """ Unpins an object and optionally runs the garbage collector """
        try:
            await self.client.pin.rm(hashRef, recursive=True)
            if rungc:
                await self.client.repo.gc()
            return True
        except aioipfs.APIError as e:
            self.debug('purge error: {}'.format(e.message))
            return False

    async def isPinned(self, hashRef):
        """
        Returns True if the IPFS object referenced by hashRef is pinned,
        False otherwise
        """
        try:
            result = await self.client.pin.ls(multihash=hashRef)
            keys = result.get('Keys', {})
            return hashRef in keys
        except aioipfs.APIError as e:
            self.debug('isPinned error: {}'.format(e.message))
            return False

    async def pinned(self, type='all'):
        """
        Returns all pinned keys of a given type
        """
        try:
            result = await self.client.pin.ls(pintype=type)
            return result.get('Keys', {})
        except aioipfs.APIError as e:
            return None

    async def pin(self, path, timeout=3600):
        async def _pin(ppath):
            async for pinStatus in self.client.pin.add(ppath):
                self.debug('Pin status: {0} {1}'.format(
                    ppath, pinStatus))
                pins = pinStatus.get('Pins', None)
                if pins is None:
                    continue
                if isinstance(pins, list) and ppath in pins:
                    # Ya estamos
                    return True
            return False
        return await self.waitFor(_pin(path), timeout)

    async def pinUpdate(self, old, new, unpin=True):
        """
        Update a pin
        """

        self.debug('pinUpdate: prev {0} new {1}'.format(old, new))
        try:
            result = await self.client.pin.update(old, new, unpin=unpin)
        except aioipfs.APIError as e:
            self.debug('pinUpdate error: {}'.format(e.message))
            return None
        else:
            self.debug('pinUpdate success: {}'.format(result))
            return result

    @async_generator
    async def list(self, path, resolve_type=True):
        """
        Lists objects in a given path and yields them
        """
        try:
            listing = await self.client.ls(path, headers=True,
                    resolve_type=resolve_type)
            objects = listing.get('Objects', [])

            for obj in objects:
                await self.sleep()
                await yield_(obj)
        except:
            pass

    async def objStat(self, path, timeout=30):
        try:
            stat = await asyncio.wait_for(self.client.object.stat(path),
                    timeout)
        except Exception as e:
            return None
        else:
            return stat

    async def objStatCtxUpdate(self, path):
        try:
            self.ctx.objectStats[path] = await self.objStat(path)
            return self.ctx.objectStats[path]
        except aioipfs.APIError as e:
            self.ctx.objectStats[path] = None

    def objStatCtxGet(self, path):
        return self.ctx.objectStats.get(path, None)

    async def addPath(self, path, recursive=True, wrap=False,
            callback=None):
        """
        Add files from path in the repo, and returns the top-level entry (the
        root directory), optionally wrapping it with a directory object

        :param str path: the path to the directory/file to import
        :param bool wrap: add a wrapping directory
        :param bool recursive: recursive import
        :return: the IPFS top-level entry
        :rtype: dict
        """
        added = None
        try:
            async for entry in self.client.add(path, quiet=True,
                    recursive=recursive,
                    wrap_with_directory=wrap):
                await self.sleep()
                added = entry
                if asyncio.iscoroutinefunction(callback):
                    await callback(entry)
        except aioipfs.APIError as err:
            self.debug(err.message)
            return None
        except Exception as e:
            self.debug('addEntry: unknown exception {}'.format(str(e)))
            return None
        else:
            return added

    async def closestPeers(self):
        peers = []
        try:
            info = await self.client.core.id()
            async for queryR in self.client.dht.query(info['ID']):
                responses = queryR.get('Responses', None)
                if not responses:
                    return None
                for resp in responses:
                    peers.append(resp['ID'])
            return peers
        except aioipfs.APIError as e:
            return None

    async def jsonLoad(self, path):
        try:
            data = await self.client.cat(path)
            return json.loads(data)
        except aioipfs.APIError as err:
            self.debug(err.message)
            return None

    async def dagPut(self, data, pin=False):
        """
        Create a new DAG object from data and returns the root CID of the DAG
        """
        dagFile = tempfile.NamedTemporaryFile()
        if not dagFile:
            return None
        try:
            dagFile.write(json.dumps(data).encode())
        except Exception as e:
            self.debug('Cannot convert DAG object: {}'.format(str(e)))
            return None

        dagFile.seek(0, 0)
        try:
            output = await self.client.dag.put(dagFile.name, pin=pin)

            if 'Cid' in output:
                return output['Cid'].get('/', None)
            else:
                self.debug('dagPut: no CID in output')
                return None
        except aioipfs.APIError as err:
            self.debug(err.message)
            return None

    async def dagGet(self, dagPath):
        """
        Get the DAG object referenced by the DAG path and returns a JSON object
        """
        try:
            output = await self.client.dag.get(dagPath)
            return json.loads(output)
        except aioipfs.APIError as err:
            self.debug(err.message)
            return None

    async def dagResolve(self, dagPath):
        """
        Resolve DAG path and return CID if path is valid and rempath empty,
        None otherwise

        Relying on RemPath is tricky and should be changed asap
        """
        try:
            resolved = await self.client.dag.resolve(dagPath)
            log.debug('dagResolve {0} {1}'.format(dagPath, resolved))
            if isDict(resolved):
                return resolved['Cid'].get('/', None) if \
                    resolved['RemPath'] == '' else None
        except aioipfs.APIError as err:
            self.debug(err.message)
            return None

    async def findProviders(self, cid, peers, verbose, nproviders):
        """
        Don't call directly, use whoProvides instead

        :param str cid: CID key to lookup
        :param list peers: a list that will be updated with the peers found
        :param int nproviders: max number of providers to look for
        """
        try:
            async for prov in self.client.dht.findprovs(cid, verbose=verbose,
                    numproviders=nproviders):
                pId = prov.get('ID', None)
                pResponses = prov.get('Responses', None)
                pType = prov.get('Type', None)

                if pType != 4: # WTF++
                    continue

                if isinstance(pResponses, list):
                    peers += pResponses
            return True
        except aioipfs.APIError as err:
            self.debug('findProviders: {}'.format(str(err.message)))
            return False

    async def whoProvides(self, key, numproviders=20, timeout=20):
        """
        Return a list of peers which provide a given key, using the DHT
        'findprovs' API call.

        Because go-ipfs's findprovs will keep looking on the dht until it
        reaches the given numproviders, we use wait_for and a timeout so that
        you always get some information about the providers.

        :param int numproviders: max number of providers to look for
        :param int timeout: timeout (seconds)
        """

        peers = []
        try:
            ret = await asyncio.wait_for(self.findProviders(key, peers, True,
                numproviders), timeout)
        except asyncio.TimeoutError as timeExc:
            # It timed out ? Return what we got
            self.debug('whoProvides: timed out')
            return peers

        return peers

    @async_generator
    async def pingWrapper(self, peer, count=3):
        """
        Peer ping async generator, used internally by the operator for specific
        ping functions.

        Items yielded are tuples of the form (time, success, text) where time
        is the latency for the ping packet, success is boolean (True if ping
        ok) and text is the ping message

        :param str peer: Peer ID
        :param int count: ping count
        """

        async for pingReply in self.client.core.ping(peer, count=count):
            if pingReply is None:
                break
            pTime = pingReply.get('Time', None)
            pSuccess = pingReply.get('Success', None)
            pText = pingReply.get('Text', '')

            self.debug('PING: {peer} ({time}) {text} {success}'.format(
                peer=peer, time=pTime, text=pText,
                success='OK' if pSuccess is True else 'ERR'))

            await yield_((pTime, pSuccess, pText.strip()))

    async def pingLowest(self, peer, count=3):
        """
        Return lowest ping latency for a peer

        :param str peer: Peer ID
        :param int count: ping count
        """

        lowest = 0
        async for time, success, text in self.pingWrapper(peer, count=count):
            if lowest == 0 and time != 0 and success is True:
                lowest = time
            if time > 0 and time < lowest and success is True:
                lowest = time
        return lowest

    async def pingAvg(self, peer, count=3):
        """
        Return average ping latency for a peer in msecs

        :param str peer: Peer ID
        :param int count: ping count
        :return: average latency in milliseconds or 0 if ping failed
        """

        received = []
        async for time, success, text in self.pingWrapper(peer, count=count):
            if time != 0 and success is True:
                received.append(time)

        if len(received) > 0:
            return float((sum(received) / len(received)) / 1000000)
        else:
            return 0

    async def versionNum(self):
        try:
            vInfo = await self.client.core.version()
            return vInfo.get('Version', None)
        except aioipfs.APIError as err:
            self.debug(err.message)
            return None

    async def hasCommand(self, name, subcmd=None):
        """
        Determines whether a given command is available on this IPFS daemon,
        and if specified, if a subcommand of this command is available.

        :param str name: command name
        :param str subcmd: subcommand name
        """
        try:
            availCmds = await self.getCommands()
            if availCmds is None:
                return False

            for command in availCmds['Subcommands']:
                if subcmd is not None:
                    for subcommand in command['Subcommands']:
                        if subcommand['Name'] == subcmd:
                            return True
                if command['Name'] == name and subcmd is None:
                    return True
            return False
        except aioipfs.APIError as err:
            self.debug(err.message)
            return False

    async def hasDagCommand(self):
        return await self.hasCommand('dag')

    async def hasP2PCommand(self):
        return await self.hasCommand('p2p')

class IPFSOpRegistry:
    _registry = {}
    _key_default = 0

    @staticmethod
    def reg(key, inst):
        if key not in IPFSOpRegistry._registry:
            IPFSOpRegistry._registry[key] = inst

    @staticmethod
    def get(key):
        return IPFSOpRegistry._registry.get(key, None)

    @staticmethod
    def regDefault(inst):
        return IPFSOpRegistry.reg(IPFSOpRegistry._key_default, inst)

    @staticmethod
    def getDefault():
        return IPFSOpRegistry._registry.get(IPFSOpRegistry._key_default)



import json
import collections
import base64

from jsonschema import validate
from jsonschema.exceptions import *

from galacteek.core.jtraverse import traverseParser
from galacteek import log

class Encoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, PubsubMessage):
            return obj.data
        if isinstance(obj, MarksBroadcastMessage):
            return obj.data
        if isinstance(obj, PeerIdentMessageV1):
            return obj.data
        return json.JSONEncoder.default(self, obj)

class PubsubMessage(collections.UserDict):
    def __init__(self, *args, **kw):
        super(PubsubMessage, self).__init__(*args, **kw)
        self.schema = None

    def __str__(self):
        return json.dumps(self.data)

    def pretty(self):
        return json.dumps(self.data, indent=4)

    @property
    def parser(self):
        return traverseParser(self.data)

    def validSchema(self, schema=None):
        sch = schema if schema else self.__class__.schema

        try:
            validate(self.data, sch)
        except ValidationError as verr:
            log.debug('Invalid JSON schema error: {}'.format(str(verr)))
            return False
        else:
            return True

    def valid(self):
        return self.validSchema()

    @staticmethod
    def make(*args, **kw):
        raise Exception('Implement static method make')

class MarksBroadcastMessage(PubsubMessage):
    TYPE = 'marksbroadcast'

    @staticmethod
    def make(marksdict):
        msg = MarksBroadcastMessage({
            'msgtype': MarksBroadcastMessage.TYPE,
            'marks': marksdict
        })
        return msg

class PeerIdentMessageV1(PubsubMessage):
    TYPE = 'peerident.v1'

    schema = {
        "title": "Peer ident",
        "description": "Peer identification message, V1",
        "type": "object",
        "properties": {
            "msgtype": {"type": "string"},
            "msg": {
                "type": "object",
                "properties": {
                    "peerid": {"type": "string"},
                    "user": {
                        "type": "object",
                        "properties": {
                            "publicdag": {
                                "type": "object",
                                "properties": {
                                    "cid" : {"type": "string"},
                                    "ipns" : {"type": "string"},
                                    "available" : {"type": "boolean"}
                                }
                            }
                        },
                    },

                    "userinfoobjref": {"type": "string"},
                    "userinfo": {
                        "type": "object",
                        "properties": {
                            "username" : {"type": "string"},
                            "firstname" : {"type": "string"},
                            "lastname" : {"type": "string"},
                            "email" : {"type": "string"},
                            "gender" : {"type": "integer"},
                            "org" : {"type": "string"},
                            "country" : {
                                "type": "object",
                                "properties": {
                                    "id": {"type": "string"},
                                    "name": {"type": "string"},
                                },
                            },
                            "date" : {"type": "object"},
                            "crypto": {"type": "object"},
                        },
                        "required": [
                            "username",
                            "firstname",
                            "lastname",
                            "email",
                            "gender",
                            "org",
                            "country"
                        ],
                    },
                },
                "required": ["peerid"]
            },
        },
    }

    @staticmethod
    def make(peerId, uInfoCid, userInfo, userDagCid, userDagIpns, p2pServices):
        msg = PeerIdentMessageV1({
            'msgtype': PeerIdentMessageV1.TYPE,
            'version': 1,
            'msg': {
                'peerid': peerId,
                'user': {
                    'publicdag': {
                        'cid': userDagCid,
                        'available': True,
                        'ipns': userDagIpns if userDagIpns else ''
                    },
                },
                'p2pservices': p2pServices,
                'userinfoobjref': uInfoCid,
            }
        })
        msg.data['msg'].update(userInfo)
        return msg

    @property
    def peer(self):
        return self.parser.traverse('msg.peerid')

    @property
    def username(self):
        return self.parser.traverse('msg.userinfo.username')

    @property
    def lastname(self):
        return self.parser.traverse('msg.userinfo.lastname')

    @property
    def country(self):
        return self.parser.traverse('msg.userinfo.country.name')

    @property
    def dateCreated(self):
        return self.parser.traverse('msg.userinfo.date.created')

    @property
    def dagCid(self):
        return self.parser.traverse('msg.user.publicdag.cid')

    @property
    def dagIpns(self):
        return self.parser.traverse('msg.user.publicdag.ipns')

    @property
    def mainPagePath(self):
        return self.parser.traverse('msg.user.mainpage')

    @property
    def userInfoObjCid(self):
        return self.parser.traverse('msg.userinfoobjref')

    @property
    def rsaPubKeyPem(self):
        return self.parser.traverse('msg.userinfo.crypto.rsa.pubkeypem')

    @property
    def msgdata(self):
        return self.data['msg']

    def valid(self):
        return self.validSchema(schema=PeerIdentMessageV1.schema)

class PeerLogoutMessage(PubsubMessage):
    TYPE = 'peerlogout'

    schema = {
        "type" : "object",
        "properties" : {
            "msgtype" : {"type" : "string"},
            "msg" : {
                "type" : "object",
                "properties" : {
                    "peerid" : {"type" : "string"},
                },
                "required": ["peerid"]
            },
        },
    }

    @staticmethod
    def make(peerid):
        msg = PeerLogoutMessage({
            'msgtype': PeerLogoutMessage.TYPE,
            'msg': {
                'peerid': peerid,
            }
        })
        return msg

    @property
    def peer(self):
        return self.parser.traverse('msg.peerid')



import pytest

from galacteek.core.ipfsmarks import *

@pytest.fixture
def bmarks(tmpdir):
    return IPFSMarks(str(tmpdir.join('bm')))

@pytest.fixture
def bmarks2(tmpdir):
    return IPFSMarks(str(tmpdir.join('bm2')))

class TestMarks:
    @pytest.mark.parametrize('path1,path2,path3', [
        (
            '/ipfs/QmT1TPVjdZ9CRnqwyQ9WygDoRgRRibFrEyWufenu92SuUV/',
            '/ipfs/QmT1TPVjdZ9CRnqwyQ9WygDoRgRRibFrEyWufenu92SuUV/a/abcdefgh/system/ak/',
            '/ipfs/Qma1TPVjdZ9CReqwyQ9Wvv3oRgRRi5FrEyWufenu92SuUV/www'
        )])
    @pytest.mark.parametrize('title', ['Broken glass'])
    def test_simpleadd(self, bmarks, path1, path2, path3, title):
        assert bmarks.add(path1, title=title, tags=['dev']) == True
        assert bmarks.add(path1, title=title) == False
        assert bmarks.add(path2, title=title) == True

        marks = bmarks.getCategoryMarks('general')
        for path, mark in marks.items():
            assert mark['metadata']['title'] == title

        assert bmarks.add(path3, title='Random', category='no/way') == True

        def deleted(m):
            mDeleted = m

        bmarks.markDeleted.connect(deleted)

        assert bmarks.delete(path2) != None
        assert bmarks.search(path2) == None

        assert bmarks.search(path1, tags=['not']) == None
        assert bmarks.search(path1, tags=['dev']) != None

        assert bmarks.add(path2, title=title) == True

        path, mark = bmarks.search(path1)
        assert mark['metadata']['title'] == title

        c = bmarks.addCategory('tests', parent=None)
        cats = bmarks.getCategories()
        assert 'general' in cats
        assert 'tests' in cats
        assert 'no/way' in cats

        bmarks.dump()

    @pytest.mark.parametrize('path',
            ['/ipfs/QmT1TPVjdZ9CRnqwyQ9WygDoRgRRibFrEyWufenu92SuUV/'])
    def test_fromjson(self, bmarks, path):
        m = IPFSHashMark.make(path, title='One')
        mark = IPFSHashMark.fromJson(m.data)

    @pytest.mark.parametrize('path1,path2,path3', [
        (
            '/ipfs/QmT1TPVjdZ9CRnqwyQ9WygDoRgRRibFrEyWufenu92SuUV/',
            '/ipfs/QmT1TPVjdZ9CRnqwyQ9WygDoRgRRibFrEyWufenu92SuUV/a/abcdefgh/system/ak/',
            '/ipfs/Qma1TPVjdZ9CReqwyQ9Wvv3oRgRRi5FrEyWufenu92SuUV/www'
        )])
    @pytest.mark.parametrize('title', ['Smooth ride'])
    def test_merge(self, bmarks, bmarks2, path1, path2, path3, title):
        bmarks.add(path1, title=title, tags=['dev'])
        bmarks2.add(path2, category='www/sub/cat', title=title)
        bmarks2.add(path3, category='www/other', title=title)
        bmarks.merge(bmarks2)

        assert bmarks.search(path2) != None
        assert bmarks.search(path3) != None

        bmarks2.add(path3, category='www/other', title=title)
        bmarks.dump()

    @pytest.mark.parametrize('feed1,mark1', [
            (
                '/ipns/QmT1TPVjdZ9CvngwyQ9WygDoRgRRibFrEyWufenu92SuUV',
                '/ipfs/QmT1TPVjdZ9CvngwyQ9WygDoRgRRibFrEyWufenu92SuUb',
                )])
    def test_follow(self, bmarks, feed1, mark1):
        fo = bmarks.follow(feed1, 'test', resolveevery=60)
        bmarks.feedAddMark(feed1, IPFSHashMark.make(mark1, title='ok'))
        assert bmarks.feedAddMark(feed1, IPFSHashMark.make(mark1, title='ok2')) == False
        bmarks.dump()

    @pytest.mark.parametrize('path1,path2', [
        (
            '/ipfs/QmT1TPVjdZ9CRnqwyQ9WygDoRgRRibFrEyWufenu92SuUV',
            '/ipfs/Qma1TPVjdZ9CReqwyQ9Wvv3oRgRRi5FrEyWufenu92SuUV/www'
        )])
    @pytest.mark.parametrize('title', ['Broken glass'])
    @pytest.mark.asyncio
    async def test_asyncquery(self, event_loop, bmarks, path1, path2, title):
        mark = IPFSHashMark.make(path1, title=title)
        mark.addTags(['one', 'two', 'three'])
        bmarks.walk(['one', 'two', 'three'])

        for i in range(0, 16):
            await bmarks.asyncQ.add('/ipfs/' + str(i),
                    category=str(i), title=title, tags=['dev'])



import os

from PyQt5.QtGui import QPixmap, QIcon
from PyQt5.QtCore import (QStandardPaths, Qt, QEvent, QObject, pyqtSignal,
        QFile)
from PyQt5.QtWidgets import (QMessageBox, QWidget, QApplication, QFileDialog,
        QInputDialog)

from . import galacteek_rc

def getIcon(iconName):
    return QIcon(QPixmap(':/share/icons/{}'.format(iconName)))

def getIconIpfsIce():
    return getIcon('ipfs-logo-128-ice.png')

def getIconIpfsWhite():
    return getIcon('ipfs-logo-128-white.png')

def getIconClipboard():
    return getIcon('clipboard.png')

def getHomePath():
    pList = QStandardPaths.standardLocations(QStandardPaths.HomeLocation)
    return pList[0] if len(pList) > 0 else os.getenv('HOME')

def messageBox(message, title=None):
    msgBox = QMessageBox()
    msgBox.setText(message)

    if title:
        msgBox.setWindowTitle(title)

    msgBox.show()
    return msgBox.exec_()

def questionBox(title, text):
    box = QMessageBox.question(None, title, text)
    return box == QMessageBox.Yes

def directorySelect(caption=''):
    return QFileDialog.getExistingDirectory(None,
        caption, getHomePath(), QFileDialog.ShowDirsOnly)

def filesSelect(filter='(*.*)'):
    result = QFileDialog.getOpenFileNames(None,
        '', getHomePath(), filter)
    if result:
        return result[0]

def saveFileSelect(filter='(*.*)'):
    result = QFileDialog.getSaveFileName(None,
        '', getHomePath(), filter)
    if result:
        return result[0]

def disconnectSig(sig, target):
    try:
        sig.disconnect(target)
    except Exception as e:
        pass

def runDialog(cls, *args, **kw):
    title = kw.pop('title', None)
    accepted = kw.pop('accepted', None)
    dlgW = cls(*args, **kw)
    def onAccept():
        if accepted:
            accepted(dlgW)
    if title:
        dlgW.setWindowTitle(title)
    if accepted:
        dlgW.accepted.connect(onAccept)
    dlgW.show()
    dlgW.exec_()
    return dlgW

def inputText(title='', label='', parent=None):
    text, ok = QInputDialog.getText(parent, title, label)
    if ok:
        return text

class IPFSTreeKeyFilter(QObject):
    deletePressed = pyqtSignal()
    copyHashPressed = pyqtSignal()
    copyPathPressed = pyqtSignal()
    explorePressed = pyqtSignal()
    returnPressed = pyqtSignal()

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyPress:
            modifiers = event.modifiers()

            key = event.key()
            if key == Qt.Key_Return:
                self.returnPressed.emit()
                return True
            if modifiers & Qt.ControlModifier:
                if key == Qt.Key_H:
                    self.copyHashPressed.emit()
                    return True
                if key == Qt.Key_P:
                    self.copyPathPressed.emit()
                    return True
                if key == Qt.Key_X:
                    self.explorePressed.emit()
                    return True
            if event.key() == Qt.Key_Delete:
                self.deletePressed.emit()
                return True
        return False

class BasicKeyFilter(QObject):
    deletePressed = pyqtSignal()
    copyPressed = pyqtSignal()
    returnPressed = pyqtSignal()

    def eventFilter(self,  obj,  event):
        if event.type() == QEvent.KeyPress:
            modifiers = event.modifiers()

            key = event.key()
            if key == Qt.Key_Return:
                self.returnPressed.emit()
            if modifiers & Qt.ControlModifier:
                if key == Qt.Key_C:
                    self.copyPressed.emit()

            if event.key() == Qt.Key_Delete:
                self.deletePressed.emit()
            return True
        return False

def clipboardSupportsSelection():
    return QApplication.clipboard().supportsSelection()

def sizeFormat(size):
    if size > (1024*1024*1024):
        return '{0:.2f} Gb'.format(size/(1024*1024*1024))
    if size > (1024*1024):
        return '{0:.2f} Mb'.format(size/(1024*1024))
    if size > 1024:
        return '{0:.2f} kb'.format(size/1024)
    if size == 0:
        return '0'
    return '{0} bytes'.format(size)



import os, os.path, re
import sys
import codecs
import subprocess
import glob
import shutil
from setuptools import setup, find_packages, Command
from distutils.command.build import build

PY_VER = sys.version_info

if PY_VER >= (3, 5):
    pass
else:
    raise RuntimeError("You need python 3.5 or newer (need async support)")

with codecs.open(os.path.join(os.path.abspath(os.path.dirname(
        __file__)), 'galacteek', '__init__.py'), 'r', 'latin1') as fp:
    try:
        version = re.findall(r"^__version__ = '([^']+)'\r?$",
                             fp.read(), re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')

def run(*args):
    p = subprocess.Popen(*args, stdout=subprocess.PIPE)
    stdout, err = p.communicate()
    return stdout

class build_docs(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from sphinx import build_main
        build_main([sys.argv[0], '-b', 'html',
            'galacteek/docs/manual/en',
            'galacteek/docs/manual/en/html'])

class build_ui(Command):
    user_options = [
        ("uiforms=", None, "UI forms list to build, separated by ','"),
    ]

    def initialize_options(self):
        self.uiforms = None

    def finalize_options(self):
        pass

    def run(self):
        uifiles = []
        uidir = 'galacteek/ui'
        dstdir = uidir

        if self.uiforms:
            uifiles = [ os.path.join(uidir, '{0}.ui'.format(form)) for form
                    in self.uiforms.split(',') ]
        else:
            uifiles = glob.iglob('{}/*.ui'.format(uidir))

        for uifile in uifiles:
            print('* Building UI form:', uifile)
            base = os.path.basename(uifile).replace('.ui', '')
            out = 'ui_{}.py'.format(base)

            run(['pyuic5', '--from-imports',
                uifile,
                '-o',
                os.path.join(uidir, out)
                ])

        run(['pylupdate5', '-verbose', 'galacteek.pro'])

        trdir = './share/translations'
        lrelease = shutil.which('lrelease-qt5')

        if not lrelease:
            lrelease = shutil.which('lrelease')

        for lang in ['en', 'fr']:
            if lrelease:
                run([lrelease,
                    os.path.join(trdir, 'galacteek_{}.ts'.format(lang)),
                    '-qm',
                    os.path.join(trdir, 'galacteek_{}.qm'.format(lang)),
                    ])
            else:
                print('lrelease was not found, cannot build translation files')

        run(['pyrcc5', os.path.join(uidir, 'galacteek.qrc'), '-o',
            os.path.join(uidir, 'galacteek_rc.py')])

class _build(build):
    sub_commands = [('build_ui', None)] + build.sub_commands

with open('README.rst', 'r') as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    install_reqs = f.read().splitlines()

setup(
    name='galacteek',
    version=version,
    license='GPL3',
    author='David Ferlier',
    author_email='galacteek@gmx.co.uk',
    url='https://gitlab.com/galacteek/galacteek',
    description='IPFS browser',
    long_description=long_description,
    include_package_data=True,
    cmdclass={'build': _build, 'build_ui': build_ui, 'build_docs': build_docs},
    packages=[
        'galacteek',
        'galacteek.docs',
        'galacteek.docs.manual',
        'galacteek.core',
        'galacteek.crypto',
        'galacteek.ipfs',
        'galacteek.ipfs.pubsub',
        'galacteek.ipfs.pb',
        'galacteek.ipfs.p2pservices',
        'galacteek.hashmarks',
        'galacteek.hashmarks.default',
        'galacteek.templates',
        'galacteek.web',
        'galacteek.ui'
    ],
    install_requires=install_reqs,
    package_data={
        'galacteek': [
             'docs/manual/en/html/*.html',
             'docs/manual/en/html/_images/*',
             'docs/manual/en/html/_static/*',
            'templates/*.html',
            'templates/public/*.html',
            'templates/public/css/*',
            'hashmarks/default/*.json'
        ]
    },
    entry_points={
        'gui_scripts': [
            'galacteek = galacteek.guientrypoint:start',
        ]
    },
    extras_require={
        'docs': [
            'sphinx>=1.7.0'
        ]
    },
    classifiers=[
        'Environment :: X11 Applications :: Qt',
        'Framework :: AsyncIO',
        'Topic :: Desktop Environment :: File Managers',
        'Topic :: Internet :: WWW/HTTP :: Browsers',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: System :: Filesystems',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords=[
        'asyncio',
        'aiohttp',
        'ipfs'
    ]
)




import sys
import json

import aioipfs
import asyncio
import socket

from galacteek import log
from galacteek.ipfs.wrappers import *
from galacteek.ipfs.multi import multiAddrTcp4

class P2PProtocol(asyncio.Protocol):
    def __init__(self):
        super(P2PProtocol, self).__init__()

        self.exitFuture = asyncio.Future()
        self.eofReceived = False

    def connection_made(self, transport):
        log.debug('Connection made')
        self.transport = transport

    def data_received(self, data):
        log.debug('Data received {}'.format(data))

    def eof_received(self):
        self.eofReceived = True

    def connection_lost(self, exc):
        self.exitFuture.set_result(True)

class P2PListener(object):
    """
    IPFS P2P listener class

    :param client: AsyncIPFS client
    :param str protocol: protocol name
    :param tuple address: address to listen on
    :param factory: protocol factory
    """

    def __init__(self, client, protocol, address, factory, loop=None):
        self.client = client
        self._protocol = protocol
        self._address = address
        self._factory = factory
        self._server = None

        self.loop = loop if loop else asyncio.get_event_loop()

    @property
    def server(self):
        return self._server

    @property
    def protocol(self):
        return self._protocol

    @property
    def address(self):
        return self._address

    @property
    def protocolFactory(self):
        return self._factory

    async def open(self):
        listenAddress = '/ip4/{addr}/tcp/{port}'.format(
            addr=self.address[0],
            port=self.address[1]
        )

        try:
            addr = await self.client.p2p.listener_open(
                self.protocol, listenAddress)
        except aioipfs.APIError as exc:
            # P2P not enabled or some other reason
            return None
        else:
            # Now that the listener is registered, create the server socket
            # and return the listener's address
            if await self.createServer():
                return addr

    async def createServer(self):
        host = self.address[0]
        for port in range(self.address[1], self.address[1]+64):
            log.debug('P2PListener: trying port {0}'.format(port))

            try:
                srv = await self.loop.create_server(self.protocolFactory,
                        host, port)
                if srv:
                    self._server = srv
                    return True
            except:
                continue


    async def close(self):
        log.debug('P2PListener: closing {0}'.format(self.protocol))
        ret = await self.client.p2p.listener_close(self.protocol)

        if self._server:
            self._server.close()
            await self._server.wait_closed()

@ipfsOpFn
async def dial(op, peer, protocol, address=None):
    loop = asyncio.get_event_loop()
    log.debug('Stream dial {0} {1}'.format(peer, protocol))
    resp = await op.client.p2p.stream_dial(peer, protocol,
            address=address)
    if resp:
        maddr = resp.get('Address', None)
        if not maddr:
            return

        ipaddr, port = multiAddrTcp4(maddr)

        if ipaddr is None or port is 0:
            return

        reader, writer = await loop.create_connection(
                lambda: P2PProtocol(), '127.0.0.1', port)



# vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
#
# Galacteek
#
# Copyright (C) 2018 David FERLIER
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

__version__ = '0.2.5'

from galacteek.core.asynclib import *
from galacteek.core import glogger

GALACTEEK_NAME = 'galacteek'

log = glogger.loggerMain



import pytest

from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtCore import QModelIndex

from galacteek.ui.modelhelpers import *

@pytest.fixture
def model():
    return QStandardItemModel()

@pytest.fixture
def stditem():
    item = QStandardItem('test')
    return item

class TestModel:
    def test_search(self, model, stditem):
        model.invisibleRootItem().appendRow(stditem)

        idxes = modelSearch(model, search='test')
        assert idxes.pop() == stditem.index()

        idxes = modelSearch(model, searchre='.*est')
        assert idxes.pop() == stditem.index()

        idxes = modelSearch(model, searchre='TEST', reflags=0)
        assert len(idxes) == 0

    def test_delete(self, model, stditem):
        model.invisibleRootItem().appendRow(stditem)
        item2 = QStandardItem('test2')
        item3 = QStandardItem('test3')
        stditem.appendRow([item2, item3])

        deleted = modelDelete(model, 'test3')
        idxes = modelSearch(model, search='test3')
        assert len(idxes) == 0




import asyncio
import functools
import time
import aiofiles

def ensure(coro, **kw):
    """ 'futcallback' should not be used in the coroutine's kwargs """
    callback = kw.pop('futcallback', None)
    future = asyncio.ensure_future(coro, **kw)
    if callback:
        future.add_done_callback(callback)
    return future

def soonish(cbk, *args, **kw):
    """ Soon. Or a bit later .. """
    loop = kw.pop('loop', asyncio.get_event_loop())
    loop.call_soon(functools.partial(cbk, *args, **kw))

class asyncify:
    def __init__(self, wrapped):
        self.wrapped = wrapped
        self.name = wrapped.__name__

    def __get__(self, inst, owner):
        def wrapper(*args, **kw):
            return asyncio.ensure_future(self.wrapped(inst, *args, **kw))
        return wrapper

def asyncifyfn(fn):
    @functools.wraps(fn)
    def ensureFn(*args, **kwargs):
        return asyncio.ensure_future(fn(*args, **kwargs))
    return ensureFn

def callAt(loop, callback, delay, *args, **kw):
    now = loop.time()
    loop.call_at(now + delay, callback, *args)

def async_enterable(f):
    """

    From aioftp.common

    Decorator. Bring coroutine result up, so it can be used as async context

    ::

        >>> async def foo():
        ...
        ...     ...
        ...     return AsyncContextInstance(...)
        ...
        ... ctx = await foo()
        ... async with ctx:
        ...
        ...     # do

    ::

        >>> @async_enterable
        ... async def foo():
        ...
        ...     ...
        ...     return AsyncContextInstance(...)
        ...
        ... async with foo() as ctx:
        ...
        ...     # do
        ...
        ... ctx = await foo()
        ... async with ctx:
        ...
        ...     # do

    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):

        class AsyncEnterableInstance:

            async def __aenter__(self):
                self.context = await f(*args, **kwargs)
                return await self.context.__aenter__()

            async def __aexit__(self, *args, **kwargs):
                await self.context.__aexit__(*args, **kwargs)

            def __await__(self):
                return f(*args, **kwargs).__await__()

        return AsyncEnterableInstance()

    return wrapper

async def asyncReadFile(path, mode='rb'):
    try:
        async with aiofiles.open(path, mode) as fd:
            return await fd.read()
    except:
        return None




import sys
import json
import os, os.path
import multiprocessing
import time
import logging
import asyncio
import re
import collections
import pkg_resources
import jinja2, jinja2.exceptions
import warnings
import concurrent.futures

from distutils.version import StrictVersion

from quamash import QEventLoop

from PyQt5.QtWidgets import QApplication, QSystemTrayIcon, QMenu
from PyQt5.QtGui import QPixmap, QIcon, QClipboard
from PyQt5.QtCore import (QCoreApplication, QUrl, QStandardPaths,
        QSettings, QTranslator, QFile, pyqtSignal, QObject,
        QTemporaryDir, QDateTime, QMessageLogger, QMimeDatabase)

from galacteek import log, soonish
from galacteek import pypicheck, GALACTEEK_NAME
from galacteek.core.asynclib import asyncify
from galacteek.core.softident import gSoftIdent
from galacteek.core.ctx import IPFSContext
from galacteek.ipfs import pinning, asyncipfsd, cidhelpers
from galacteek.ipfs.ipfsops import *
from galacteek.ipfs.wrappers import *
from galacteek.ipfs.pubsub import *
from galacteek.ipfs.feeds import FeedFollower

from galacteek.ui import mainui, galacteek_rc, downloads, browser, hashmarks
from galacteek.ui.helpers import *
from galacteek.ui.i18n import *

from galacteek.appsettings import *
from galacteek.core.ipfsmarks import IPFSMarks

from yarl import URL

import aioipfs

# Application's i18n messages

# IPFS daemon messages

def iIpfsDaemonStarted():
    return QCoreApplication.translate('Galacteek', 'IPFS daemon started')

def iIpfsDaemonGwStarted():
    return QCoreApplication.translate('Galacteek',
        "IPFS daemon's gateway started")

def iIpfsDaemonReady():
    return QCoreApplication.translate('Galacteek', 'IPFS daemon is ready')

def iIpfsDaemonProblem():
    return QCoreApplication.translate('Galacteek',
        'Problem starting IPFS daemon')

def iIpfsDaemonInitProblem():
    return QCoreApplication.translate('Galacteek',
        'Problem initializing the IPFS daemon (check the ports configuration)')

def iIpfsDaemonWaiting(count):
    return QCoreApplication.translate('Galacteek',
        'IPFS daemon: waiting for connection (try {0})'.format(count))

class IPFSConnectionParams(object):
    def __init__(self, host, apiport, gwport):
        self._host = host
        self._apiPort = apiport
        self._gatewayPort = gwport

        self._gatewayUrl = URL.build(host=self.host,
            port=self.gatewayPort, scheme='http', path='')

    @property
    def host(self):
        return self._host

    @property
    def apiPort(self):
        return self._apiPort

    @property
    def gatewayPort(self):
        return self._gatewayPort

    @property
    def gatewayUrl(self):
        return self._gatewayUrl

def appOperator():
    app = QApplication.instance()
    if not app:
        raise Exception('No Qt Application running')
    return app.ipfsOpMain

class gIpfsOpMain(ipfsClassW):
    def __get__(self, inst, owner):
        async def wrapper(*args, **kw):
            op = appOperator()
            if op:
                return await self.wrapped(inst, op, *args, **kw)
        return wrapper

class GalacteekApplication(QApplication):
    """
    Galacteek application class

    :param bool debug: enable debugging
    :param str profile: application profile
    """

    manualAvailable = pyqtSignal(str, dict)

    def __init__(self, debug=False, profile='main'):
        QApplication.__init__(self, sys.argv)

        QCoreApplication.setApplicationName(GALACTEEK_NAME)

        self._appProfile = profile
        self._loop = None
        self._executor = None
        self._ipfsClient = None
        self._ipfsOpMain = None
        self._ipfsd = None

        self.translator = None

        self.mainWindow = None
        self.gWindows = []

        self.feedFollowerTask = None

        self._debugEnabled = debug

        self.ipfsCtx = IPFSContext(self)

        self.setupPaths()
        self.setupClipboard()
        self.setupSchemeHandlers()

        self.initSettings()
        self.setupTranslator()
        self.initSystemTray()
        self.initMisc()
        self.createMainWindow()

        self.setStyle()
        self.clipboardInit()

    @property
    def debugEnabled(self):
        return self._debugEnabled

    @property
    def appProfile(self):
        return self._appProfile

    @property
    def ipfsd(self):
        return self._ipfsd

    @property
    def loop(self):
        return self._loop

    @property
    def executor(self):
        return self._executor

    @loop.setter
    def loop(self, newLoop):
        self._loop = newLoop

    @property
    def allTasks(self):
        return asyncio.Task.all_tasks(loop=self.loop)

    @property
    def pendingTasks(self):
        return [task for task in asyncio.Task.all_tasks(loop=self.loop) if \
                not task.done()]

    @property
    def ipfsClient(self):
        return self._ipfsClient

    @ipfsClient.setter
    def ipfsClient(self, client):
        self.debug('IPFS client changed: {}'.format(client))
        self._ipfsClient = client

    @property
    def ipfsOpMain(self):
        return self._ipfsOpMain

    @ipfsOpMain.setter
    def ipfsOpMain(self, op):
        """ The main IPFS operator, used by @ipfsOp """
        self.debug('Main IPFS operator upgrade: ID {}'.format(op.uid))
        self._ipfsOpMain = op

    @property
    def gatewayAuthority(self):
        params = self.getIpfsConnectionParams()
        return '{0}:{1}'.format(params.host, params.gatewayPort)

    @property
    def gatewayUrl(self):
        params = self.getIpfsConnectionParams()
        return params.gatewayUrl

    @property
    def ipfsBinLocation(self):
        return self._ipfsBinLocation

    @property
    def ipfsDataLocation(self):
        return self._ipfsDataLocation

    def setStyle(self):
        qssPath = ":/share/static/qss/galacteek.qss"
        qssFile = QFile(qssPath)

        try:
            qssFile.open(QFile.ReadOnly)
            styleSheetBa = qssFile.readAll()
            styleSheetStr = styleSheetBa.data().decode('utf-8')
            self.setStyleSheet(styleSheetStr)
        except:
            # that would probably occur if the QSS is not in the resources file..
            # set some default stylesheet here?
            pass

    def debug(self, msg):
        if self.debugEnabled:
            log.debug(msg)

    def initSystemTray(self):
        self.systemTray = QSystemTrayIcon(self)
        self.systemTray.setIcon(getIcon('ipfs-logo-128-ice-text.png'))
        self.systemTray.show()
        self.systemTray.activated.connect(self.onSystemTrayIconClicked)
        self.systemTray.setToolTip('Galacteek')

        systemTrayMenu = QMenu(None)

        action = systemTrayMenu.addAction('Quit')
        action.setIcon(QIcon.fromTheme('application-exit'))
        action.triggered.connect(self.onExit)

        self.systemTray.setContextMenu(systemTrayMenu)

    def initMisc(self):
        self.jinjaEnv = jinja2.Environment(
            loader=jinja2.PackageLoader('galacteek', 'templates'))

        self.manuals = ManualsImporter(self)
        self.mimeDb = QMimeDatabase()

        self.downloadsManager = downloads.DownloadsManager(self)
        self.marksLocal = IPFSMarks(self.localMarksFileLocation)
        self.importDefaultHashmarks(self.marksLocal)

        self.marksLocal.addCategory('general')
        self.marksNetwork = IPFSMarks(self.networkMarksFileLocation,
                autosave=False)

        self.tempDir = QTemporaryDir()
        if not self.tempDir.isValid():
            pass

    def importDefaultHashmarks(self, marksLocal):
        pkg = 'galacteek.hashmarks.default'
        try:
            listing = pkg_resources.resource_listdir(pkg, '')
            for fn in listing:
                if fn.endswith('.json'):
                    path = pkg_resources.resource_filename(pkg, fn)
                    marks = IPFSMarks(path)
                    marksLocal.merge(marks)
        except Exception as e:
            self.debug(str(e))

    def setupTranslator(self):
        if self.translator:
            QApplication.removeTranslator(self.translator)

        self.translator = QTranslator()
        QApplication.installTranslator(self.translator)
        lang = self.settingsMgr.getSetting(CFG_SECTION_UI, CFG_KEY_LANG)
        self.translator.load(':/share/translations/galacteek_{0}.qm'.format(
            lang))

    def createMainWindow(self, show=True):
        self.mainWindow = mainui.MainWindow(self)
        if show is True:
            self.mainWindow.show()

    def onSystemTrayIconClicked(self, reason):
        if reason == QSystemTrayIcon.Unknown:
            pass
        elif reason == QSystemTrayIcon.Context:
            pass
        elif reason == QSystemTrayIcon.DoubleClick:
            self.mainWindow.show()
        else:
            pass

    def systemTrayMessage(self, title, message, timeout=2000,
            messageIcon=QSystemTrayIcon.Information):
        self.systemTray.showMessage(title, message, messageIcon, timeout)

    @ipfsOp
    async def setupRepository(self, op):
        pubsubEnabled = True # mandatory now ..
        hExchEnabled = self.settingsMgr.isTrue(CFG_SECTION_IPFS,
            CFG_KEY_HASHMARKSEXCH)

        self.ipfsCtx.resources['ipfs-logo-ice'] = await self.importQtResource(
               '/share/icons/ipfs-logo-128-ice.png')
        self.ipfsCtx.resources['ipfs-cube-64'] = await self.importQtResource(
                '/share/icons/ipfs-cube-64.png')
        self.ipfsCtx.resources['ipfs-logo-text'] = await self.importQtResource(
                '/share/icons/ipfs-logo-text-128-ice-white.png')

        await self.ipfsCtx.setup(pubsubEnable=pubsubEnabled,
                pubsubHashmarksExch=hExchEnabled)
        await self.ipfsCtx.profilesInit()

        ensure(self.manuals.importManualMain())

        self.feedFollower = FeedFollower(self, self.marksLocal)
        self.feedFollowerTask = self.task(self.feedFollower.process)

        self.loop.call_soon(self.ipfsCtx.ipfsRepositoryReady.emit)

    @ipfsOp
    async def importQtResource(self, op, path):
        rscFile = QFile(':{0}'.format(path))

        try:
            rscFile.open(QFile.ReadOnly)
            data = rscFile.readAll().data()
            entry = await op.client.add_bytes(data)
        except Exception as e:
            log.debug('importQtResource: {}'.format(str(e)))
        else:
            return entry

    def ipfsTask(self, fn, *args, **kw):
        """ Schedule an async IPFS task """
        return self.loop.create_task(fn(self.ipfsClient,
            *args, **kw))

    def ipfsTaskOp(self, fn, *args, **kw):
        """ Schedule an async IPFS task using an IPFSOperator instance """
        client = self.ipfsClient
        if client:
            return self.loop.create_task(fn(
                self.getIpfsOperator(), *args, **kw))

    def getIpfsOperator(self):
        """ Returns a new IPFSOperator with the currently active IPFS client"""
        return IPFSOperator(self.ipfsClient, ctx=self.ipfsCtx,
            debug=self.debugEnabled)

    def getIpfsConnectionParams(self):
        mgr = self.settingsMgr

        section = CFG_SECTION_IPFSD
        if mgr.isTrue(section, CFG_KEY_ENABLED):
            return IPFSConnectionParams('localhost',
                    mgr.getSetting(section, CFG_KEY_APIPORT),
                    mgr.getSetting(section, CFG_KEY_HTTPGWPORT)
            )
        else:
            section = CFG_SECTION_IPFSCONN1
            return IPFSConnectionParams(
                    mgr.getSetting(section, CFG_KEY_HOST),
                    mgr.getSetting(section, CFG_KEY_APIPORT),
                    mgr.getSetting(section, CFG_KEY_HTTPGWPORT)
            )

    async def updateIpfsClient(self):
        connParams = self.getIpfsConnectionParams()
        client = aioipfs.AsyncIPFS(host=connParams.host,
                port=connParams.apiPort, loop=self.loop)
        self.ipfsClient = client
        self.ipfsCtx.ipfsClient = client
        self.ipfsOpMain = self.getIpfsOperator()

        IPFSOpRegistry.regDefault(self.ipfsOpMain)

        self.loop.call_soon(self.ipfsCtx.ipfsConnectionReady.emit)

        await self.setupRepository()

    async def stopIpfsServices(self):
        await self.ipfsCtx.shutdown()

        if self.feedFollowerTask is not None:
            self.feedFollowerTask.cancel()

    def setupAsyncLoop(self):
        """
        Install the quamash event loop and enable debugging
        """

        loop = QEventLoop(self)
        asyncio.set_event_loop(loop)
        logging.getLogger('quamash').setLevel(logging.INFO)

        if self.debugEnabled:
            logging.getLogger('asyncio').setLevel(logging.DEBUG)
            loop.set_debug(True)
            warnings.simplefilter('always', ResourceWarning)
            warnings.simplefilter('always', BytesWarning)
            warnings.simplefilter('always', ImportWarning)

        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=5)

        self.loop = loop
        return loop

    def task(self, fn, *args, **kw):
        return self.loop.create_task(fn(*args, **kw))

    def setupPaths(self):
        qtDataLocation = QStandardPaths.writableLocation(QStandardPaths.DataLocation)

        if not qtDataLocation:
            raise Exception('No writable data location found')

        self._dataLocation = os.path.join(
                qtDataLocation, self._appProfile)

        self._ipfsBinLocation = os.path.join(qtDataLocation, 'ipfs-bin')
        self._ipfsDataLocation = os.path.join(self._dataLocation, 'ipfs')
        self.marksDataLocation = os.path.join(self._dataLocation, 'marks')
        self.cryptoDataLocation = os.path.join(self._dataLocation, 'crypto')
        self.gpgDataLocation = os.path.join(self.cryptoDataLocation, 'gpg')
        self.localMarksFileLocation = os.path.join(self.marksDataLocation,
            'ipfsmarks.local.json')
        self.networkMarksFileLocation = os.path.join(self.marksDataLocation,
            'ipfsmarks.network.json')

        qtConfigLocation = QStandardPaths.writableLocation(
                QStandardPaths.ConfigLocation)
        self.configDirLocation = os.path.join(
                qtConfigLocation, GALACTEEK_NAME, self._appProfile)
        self.settingsFileLocation = os.path.join(
                self.configDirLocation, '{}.conf'.format(GALACTEEK_NAME))

        for dir in [ self._ipfsDataLocation,
                self.ipfsBinLocation,
                self.marksDataLocation,
                self.cryptoDataLocation,
                self.gpgDataLocation,
                self.configDirLocation ]:
            if not os.path.exists(dir):
                os.makedirs(dir)

        self.defaultDownloadsLocation = QStandardPaths.writableLocation(
            QStandardPaths.DownloadLocation)

        self.debug('Datapath: {0}, config: {1}, configfile: {2}'.format(
                self._dataLocation,
                self.configDirLocation,
                self.settingsFileLocation))

        os.environ['PATH'] += os.pathsep + self.ipfsBinLocation

    def initSettings(self):
        self.settingsMgr = SettingsManager(path=self.settingsFileLocation)
        setDefaultSettings(self)
        self.settingsMgr.sync()

    def startIpfsDaemon(self, goIpfsPath='ipfs', migrateRepo=False):
        if self.ipfsd is not None: # we only support one daemon for now
            return

        #pubsubEnabled = self.settingsMgr.isTrue(CFG_SECTION_IPFS,
        #    CFG_KEY_PUBSUB)
        pubsubEnabled = True # mandatory now ..
        corsEnabled = self.settingsMgr.isTrue(CFG_SECTION_IPFSD,
            CFG_KEY_CORS)

        sManager = self.settingsMgr
        section = CFG_SECTION_IPFSD

        # Instantiate an IPFS daemon using asyncipfsd and
        # start it in a task, monitoring the initialization process

        self._ipfsd = asyncipfsd.AsyncIPFSDaemon(
            self.ipfsDataLocation, goIpfsPath=goIpfsPath,
            apiport=sManager.getInt(section, CFG_KEY_APIPORT),
            swarmport=sManager.getInt(section, CFG_KEY_SWARMPORT),
            gatewayport=sManager.getInt(section, CFG_KEY_HTTPGWPORT),
            swarmLowWater=sManager.getInt(section, CFG_KEY_SWARMLOWWATER),
            swarmHighWater=sManager.getInt(section, CFG_KEY_SWARMHIGHWATER),
            storageMax=sManager.getInt(section, CFG_KEY_STORAGEMAX),
            pubsubEnable=pubsubEnabled, corsEnable=corsEnabled,
            migrateRepo=migrateRepo, debug=self.debug,
            loop=self.loop)

        self.task(self.startIpfsdTask, self.ipfsd)

    async def startIpfsdTask(self, ipfsd):
        started = await ipfsd.start()
        self.mainWindow.statusMessage(iIpfsDaemonStarted())

        if started is False:
            return self.systemTrayMessage('IPFS', iIpfsDaemonProblem())

        running = False

        self.systemTrayMessage('IPFS', iIpfsDaemonStarted(),
            timeout=1000)

        # Use asyncio.wait_for to wait for the proto.eventStarted
        # event to be fired. 

        for attempt in range(1, 32):
            self.mainWindow.statusMessage(iIpfsDaemonWaiting(attempt))
            try:
                await asyncio.wait_for(ipfsd.proto.eventStarted.wait(), 1)
            except asyncio.TimeoutError as err:
                # Event not set yet, wait again
                log.debug('IPFSD: timeout occured while waiting for daemon to \
                        start (attempt: {0})'.format(attempt))
                continue
            else:
                # Event was set, good to go
                self.systemTrayMessage('IPFS',
                    iIpfsDaemonReady(), timeout=2000)
                running = True
                break

        if running is True:
            ensure(self.updateIpfsClient())
        else:
            self.systemTrayMessage('IPFS', iIpfsDaemonInitProblem())

    def setupClipboard(self):
        self.appClipboard = self.clipboard()
        self.clipTracker = ClipboardTracker(self.appClipboard)

    def clipboardInit(self):
        self.clipTracker.clipboardInit()

    def setClipboardText(self, text):
        self.clipTracker.setText(text)

    def getClipboardText(self):
        return self.clipTracker.getText()

    def setupSchemeHandlers(self):
        self.ipfsSchemeHandler = browser.IPFSSchemeHandler(self)

    def subUrl(self, path):
        """ Joins the gatewayUrl and path to form a new URL """
        sub = QUrl(str(self.gatewayUrl))
        sub.setPath(path)
        return sub

    def getJinjaTemplate(self, name):
        try:
            tmpl = self.jinjaEnv.get_template(name)
        except jinja2.exceptions.TemplateNotFound as e:
            return None
        else:
            return tmpl

    @asyncify
    async def checkReleases(self):
        self.debug('Checking for new releases')
        newR = await pypicheck.newReleaseAvailable()
        if newR:
            self.systemTrayMessage('Galacteek',
                iNewReleaseAvailable(), timeout=8000)

    def showTasks(self):
        for task in self.pendingTasks:
            self.debug('Pending task: {}'.format(task))

    @asyncify
    async def onExit(self):
        await self.exitApp()

    async def exitApp(self):
        await self.stopIpfsServices()

        if self.ipfsd:
            self.ipfsd.stop()

        if self.debug:
            self.showTasks()

        self.tempDir.remove()
        self.quit()

class ClipboardTracker(QObject):
    """
    Tracks the system's clipboard activity and emits signals
    depending on whether or not the clipboard contains an IPFS CID or path
    """
    clipboardHasIpfs = pyqtSignal(bool, str, str)
    clipboardHistoryChanged = pyqtSignal(dict)

    def __init__(self, clipboard):
        super(ClipboardTracker, self).__init__()

        self.hasIpfs = False
        self.history = {}
        self.clipboard = clipboard
        self.clipboard.changed.connect(self.onClipboardChanged)
        self.clipboardHasIpfs.connect(self.onHasIpfs)

    def onClipboardChanged(self, mode):
        text = self.clipboard.text(mode)
        self.clipboardProcess(text, clipboardMode=mode)

    def clipboardProcess(self, text, clipboardMode=None):
        """
        Process the contents of the clipboard. If it is a valid CID/path, emit a
        signal, processed by the main window for the clipboard loader button
        """
        if not text or len(text) > 1024: # that shouldn't be worth handling
            return

        text = text.strip()
        ma = cidhelpers.ipfsRegSearchPath(text)

        if ma:
            # The clipboard contains a full IPFS path
            cid = ma.group('cid')
            if not cidhelpers.cidValid(cid):
                return
            path = ma.group('fullpath')
            self.hRecord(path)
            return self.clipboardHasIpfs.emit(True, cid, path)

        ma = cidhelpers.ipnsRegSearchPath(text)
        if ma:
            # The clipboard contains a full IPNS path
            path = ma.group('fullpath')
            self.hRecord(path)
            return self.clipboardHasIpfs.emit(True, None, path)

        ma = cidhelpers.ipfsRegSearchCid(text)
        if ma:
            # The clipboard simply contains a CID
            cid = ma.group('cid')
            if not cidhelpers.cidValid(cid):
                return

            path = joinIpfs(cid)
            self.hRecord(path)
            return self.clipboardHasIpfs.emit(True, cid, path)

        # Not a CID/path
        if clipboardMode == self.clipboardPreferredMode():
            self.clipboardHasIpfs.emit(False, None, None)

    def hLookup(self, path):
        for hTs, hItem in self.history.items():
            if hItem['path'] == path:
                return hTs, hItem

    def hRecord(self, path):
        """ Records an item in the history and emits a signal """
        now = time.time()
        itLookup = self.hLookup(path)
        self.history[now] = {
            'path': path,
            'date': QDateTime.currentDateTime()
        }
        self.clipboardHistoryChanged.emit(self.getHistory())

    def getHistory(self):
        return collections.OrderedDict(sorted(self.history.items(),
            key=lambda t: t[0]))

    def clearHistory(self):
        self.history = {}
        self.clipboardHistoryChanged.emit(self.getHistory())

    def getHistoryLatest(self):
        """ Returns latest history item """
        h = self.getHistory()
        try:
            return h.popitem(last=True)[1]
        except KeyError:
            return None

    def getCurrent(self):
        """ Returns current clipboard item """
        if self.hasIpfs:
            return self.getHistoryLatest()

    def clipboardInit(self):
        """ Used to process the clipboard's content on application's init """
        text = self.getText()
        self.clipboardProcess(text)

    def clipboardPreferredMode(self):
        return QClipboard.Selection if self.clipboard.supportsSelection() \
                else QClipboard.Clipboard

    def setText(self, text):
        """ Sets the clipboard's text content """
        self.clipboard.setText(text, self.clipboardPreferredMode())

    def getText(self):
        """ Returns clipboard's text content from the preferred source """
        return self.clipboard.text(self.clipboardPreferredMode())

    def onHasIpfs(self, valid, cid, path):
        self.hasIpfs = valid

class ManualsImporter(QObject):
    """ Imports the HTML manuals in IPFS """

    def __init__(self, app):
        super(ManualsImporter, self).__init__()

        self.app = app
        self.registry = {}

    async def importManualMain(self):
        try:
            listing = pkg_resources.resource_listdir(
                    'galacteek.docs.manual', '')
            for entry in listing:
                if entry.startswith('__'):
                    continue
                await self.importManualLang(entry)
        except Exception as e:
            log.debug('Failed importing manuals {0}'.format(str(e)))

    async def importManualLang(self, lang):
        try:
            docPath = pkg_resources.resource_filename('galacteek.docs.manual',
                '{0}/html'.format(lang))
            await self.importDocPath(docPath, lang)
        except Exception as e:
            log.debug('Failed importing manual ({0}) {1}'.format(
                lang, str(e)))

    @ipfsOp
    async def importDocPath(self, ipfsop, docPath, lang):
        docEntry = await ipfsop.addPath(docPath)
        if docEntry:
            self.registry[lang] = docEntry
            self.app.manualAvailable.emit(lang, docEntry)




from async_generator import async_generator, yield_

import aiohttp
import aiofiles
import sys
import os, os.path
import platform
import tarfile, zipfile
import tempfile
import asyncio
import shutil

@async_generator
async def distIpfsExtract(dstdir='.', software='go-ipfs', executable='ipfs',
        site='dist.ipfs.io', version='0.4.17', loop=None):

    """ Fetch a distribution archive from dist.ipfs.io and extracts the
        wanted executable to dstdir. Yields progress messages """

    arch = None
    osType = None
    ext  = None
    exeExt = ''

    pArch = platform.architecture()
    pMachine = platform.machine()
    pSystem = platform.system()

    if pMachine == 'x86_64':
        arch = 'amd64'
    elif pMachine == 'i386':
        arch = '386'

    if pSystem == 'Linux':
        osType = 'linux'
        arExt = '.tar.gz'
    elif pSystem == 'Windows':
        osType = 'windows'
        arExt = '.zip'
        exeExt = '.exe'
    elif pSystem == 'FreeBSD':
        osType = 'freebsd'
        arExt = '.tar.gz'
    elif pSystem == 'Darwin':
        osType = 'darwin'
        arExt = '.tar.gz'

    if arch is None or osType is None or arExt is None:
        return False

    tmpDst = tempfile.mkdtemp(prefix='distipfs')
    if not tmpDst:
        return False

    fileName = '{software}_v{version}_{os}-{arch}{ext}'.format(
        software=software, version=version, arch=arch, os=osType, ext=arExt)

    url = 'https://{site}/{software}/v{version}/{filename}'.format(
            software=software, site=site, version=version, arch=arch,
            os=osType, ext=arExt, filename=fileName)

    tmpFile = tempfile.NamedTemporaryFile(suffix=arExt, delete=False)
    arPath = tmpFile.name

    def statusMessage(code, msg):
        return ((code, '{0}: {1}'.format(fileName, msg)))

    await yield_((0, 'Starting download: {0} ...'.format(url)))

    async with aiofiles.open(arPath, 'w+b') as fd:
        bytesRead = 0
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as resp:
                if resp.status == 404:
                    await yield_(statusMessage(0,
                        'Error downloading (file not found)'))
                    return False
                else:
                    await yield_(statusMessage(0, 'File found!'))
                while True:
                    data = await resp.content.read(262144)
                    if not data:
                        break
                    await fd.write(data)
                    bytesRead += len(data)
                    await yield_(statusMessage(0,
                        'received {} bytes'.format(bytesRead)))
                    await asyncio.sleep(0)

    def extract(path, dest):
        fname = os.path.basename(path)
        if fname.endswith('.zip'):
            opener, mode = zipfile.ZipFile, 'r'
        elif fname.endswith('.tar.gz') or path.endswith('.tgz'):
            opener, mode = tarfile.open, 'r:gz'
        else:
            # can't handle that
            return False
        try:
            # the path inside the archive.
            execPath = os.path.join(software, executable + exeExt)
            with opener(path, mode=mode) as tf:
                tf.extract(execPath, path=dest)
                shutil.copy(os.path.join(dest, execPath) , dstdir)
            os.unlink(arPath)
            shutil.rmtree(dest, ignore_errors=True)
            return True
        except Exception as e:
            print('Could not extract archive file:', str(e), file=sys.stderr)
            os.unlink(arPath)
            return False

    await yield_(statusMessage(0, 'Extracting distribution ..'))
    loop = loop if loop else asyncio.get_event_loop()
    tarF = loop.run_in_executor(None, extract, arPath, tmpDst)
    return await tarF



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/addfeeddialog.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_AddFeedDialog(object):
    def setupUi(self, AddFeedDialog):
        AddFeedDialog.setObjectName("AddFeedDialog")
        AddFeedDialog.resize(438, 181)
        self.verticalLayout = QtWidgets.QVBoxLayout(AddFeedDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.resourceLabel = QtWidgets.QLabel(AddFeedDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.resourceLabel.setFont(font)
        self.resourceLabel.setText("")
        self.resourceLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.resourceLabel.setObjectName("resourceLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.resourceLabel)
        self.label = QtWidgets.QLabel(AddFeedDialog)
        self.label.setObjectName("label")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label)
        self.feedName = QtWidgets.QLineEdit(AddFeedDialog)
        self.feedName.setObjectName("feedName")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.feedName)
        self.label_2 = QtWidgets.QLabel(AddFeedDialog)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.resolve = QtWidgets.QSpinBox(AddFeedDialog)
        self.resolve.setMinimum(20)
        self.resolve.setMaximum(2592000)
        self.resolve.setProperty("value", 3600)
        self.resolve.setObjectName("resolve")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.resolve)
        self.share = QtWidgets.QCheckBox(AddFeedDialog)
        self.share.setObjectName("share")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.share)
        self.buttonBox = QtWidgets.QDialogButtonBox(AddFeedDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.buttonBox)
        self.autoPin = QtWidgets.QCheckBox(AddFeedDialog)
        self.autoPin.setChecked(False)
        self.autoPin.setObjectName("autoPin")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.autoPin)
        self.verticalLayout.addLayout(self.formLayout)

        self.retranslateUi(AddFeedDialog)
        self.buttonBox.accepted.connect(AddFeedDialog.accept)
        self.buttonBox.rejected.connect(AddFeedDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(AddFeedDialog)

    def retranslateUi(self, AddFeedDialog):
        _translate = QtCore.QCoreApplication.translate
        AddFeedDialog.setWindowTitle(_translate("AddFeedDialog", "Dialog"))
        self.label.setText(_translate("AddFeedDialog", "Feed name"))
        self.label_2.setText(_translate("AddFeedDialog", "Resolve frequency (seconds)"))
        self.share.setText(_translate("AddFeedDialog", "Share"))
        self.autoPin.setText(_translate("AddFeedDialog", "Auto PIN"))




import collections

class IPFSEntryCache(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self._cache = collections.OrderedDict()

    def register(self, entry):
        if entry['Hash'] not in self._cache:
            self._cache[entry['Hash']] = entry

    def __contains__(self, oHash):
        return oHash in self._cache

    def purge(self, oHash):
        if oHash in self:
            del self._cache[oHash]

    def getByType(self, etype):
        entries = {}
        for ohash, entry in self._cache.items():
            if entry['Type'] == etype:
                entries[ohash] = entry
        return entries



from async_generator import async_generator, yield_, yield_from_
import asyncio
import re

from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtCore import QModelIndex

""" Helper functions to operate on QT item models """

def modelWalk(model, parent=QModelIndex(), search=None,
        searchre=None, columns=None, delete=False, dump=False,
        reflags=re.IGNORECASE, maxdepth=0, depth=0):
    """Walk over every model index in the given model looking for data matching
    the search parameter (or searchre if you want to use a regexp)

    Yields matching model indexes

    :param QAbstractItemModel model: the QT model
    :param QModelIndex parent: the parent item index in the model
    :param list columns: integer list of columns to search
    :param str search: string to search
    :param str searchre: regexp to search
    :param int reflags: regexp flags passed to re.search
    :param bool delete: delete matching indexes
    :param int maxdepth: maximum tree depth
    :return: list of matching indexes
    """

    for row in range(0, model.rowCount(parent)):
        index0 = model.index(row, 0, parent)
        if columns:
            cols = columns
        else:
            cols = [c for c in range(0, model.columnCount(parent))]
        for col in cols:
            index = model.index(row, col, parent)
            if not index:
                continue
            data = model.data(index)
            if search:
                if search == str(data):
                    if delete == True:
                        model.removeRow(row, parent)
                        model.submit()
                    yield index
            if searchre:
                if re.search(searchre, str(data), reflags):
                    yield index
        if model.hasChildren(index0):
            if maxdepth == 0 or (maxdepth > depth):
                yield from modelWalk(model, parent=index0,
                        search=search, columns=columns, delete=delete,
                        dump=dump, searchre=searchre, reflags=reflags,
                        maxdepth=maxdepth, depth=depth)
                depth += 1

def modelSearch(model, parent=QModelIndex(), columns=None, delete=False,
        search=None, searchre=None, reflags=re.IGNORECASE):
    """Searches data in a QT model, see modelWalk for more info on
    parameters"""
    return list(modelWalk(model, parent=parent, columns=columns,
        search=search, delete=delete, searchre=searchre,
        reflags=reflags))

def modelDelete(model, search):
    """Delete items matching search from a QT model """
    return list(modelSearch(model, search=search, delete=True))

""" Asynchronous equivalents for when calling from coroutines """

@async_generator
async def modelWalkAsync(model, parent=QModelIndex(), search=None,
        columns=None, delete=False, dump=False,
        maxdepth=0, depth=0):
    for row in range(0, model.rowCount(parent)):
        await asyncio.sleep(0)

        index0 = model.index(row, 0, parent)
        if columns:
            cols = columns
        else:
            cols = [c for c in range(0, model.columnCount(parent))]
        for col in cols:
            index = model.index(row, col, parent)
            if not index:
                continue
            data = model.data(index)
            if search:
                if search == data:
                    if delete == True:
                        model.removeRow(row, parent)
                        model.submit()
                    await yield_(index)
        await asyncio.sleep(0)
        if model.hasChildren(index0):
            if maxdepth == 0 or (maxdepth > depth):
                await yield_from_( modelWalkAsync(model, parent=index0,
                    search=search, columns=columns, delete=delete,
                    dump=dump, maxdepth=maxdepth, depth=depth) )
                depth += 1

async def modelSearchAsync(model, parent=QModelIndex(), search=None,
        columns=None, delete=False, maxdepth=0, depth=0):
    items = []
    async for v in modelWalkAsync(model, parent=parent, columns=columns,
            search=search, delete=delete, maxdepth=maxdepth, depth=depth):
        items.append(v)
    return items

async def modelDeleteAsync(model, search):
    return await modelSearchAsync(model, search=search, delete=True)

class UneditableItem(QStandardItem):
    def __init__(self, text, icon=None):
        if icon:
            super(UneditableItem, self).__init__(icon, text)
        else:
            super(UneditableItem, self).__init__(text)
        self.setEditable(False)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/mediaplayer.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MediaPlayer(object):
    def setupUi(self, MediaPlayer):
        MediaPlayer.setObjectName("MediaPlayer")
        MediaPlayer.resize(400, 300)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(MediaPlayer)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_2.addLayout(self.verticalLayout)

        self.retranslateUi(MediaPlayer)
        QtCore.QMetaObject.connectSlotsByName(MediaPlayer)

    def retranslateUi(self, MediaPlayer):
        _translate = QtCore.QCoreApplication.translate
        MediaPlayer.setWindowTitle(_translate("MediaPlayer", "Form"))





import os.path
import re
import mimetypes
import asyncio

from html.parser import HTMLParser

import aioipfs

class TitleParser(HTMLParser):
    def __init__(self):
        super().__init__()

        self.inHead = False
        self.inTitle = False
        self.data = ''
        self.title = None

    def handle_starttag(self, tag, attrs):
        self.data = ''
        if tag == 'head':
            self.inHead = True
        if tag == 'title' and self.inHead is True:
            self.inTitle = True

    def handle_endtag(self, tag):
        if tag == 'head':
            self.inHead = False
        if tag == 'title' and self.inHead is True:
            self.inTitle = False
            self.title = self.data

    def handle_data(self, data):
        self.data += data

    def getTitle(self):
        return self.title

def runTitleParser(data):
    try:
        parser = TitleParser()
        parser.feed(data.decode())
        return parser.getTitle()
    except:
        return None

async def getTitleDirectory(client, path):
    indexPath = os.path.join(path, 'index.html')

    try:
        data = await client.cat(indexPath)
    except aioipfs.APIError as exc:
        if exc.code == 0 and exc.message.startswith('no link named'):
            return None
        if exc.code == 0 and exc.message == 'this dag node is a directory':
            # Wicked!
            return None
    else:
        return runTitleParser(data)

async def getTitle(client, path):
    """
    Lookup IPFS object with path and if it's an HTML document, return its title
    If path is a directory, call getTitleDirectory() which will return the
    title of the index.html file inside that directory if found
    """
    try:
        data = await client.cat(path)
    except aioipfs.APIError as exc:
        if exc.code == 0 and exc.message.startswith('no link named'):
            return None
        if exc.code == 0 and exc.message == 'this dag node is a directory':
            return await getTitleDirectory(client, path)
    else:
        return runTitleParser(data)




import collections
import os.path
import json
import sys

from PyQt5.QtCore import pyqtSignal, QObject

class QJSONObj(QObject):
    changed = pyqtSignal()

    def __init__(self, data=None, **kw):
        super().__init__()

        self.__dict__.update(kw)
        self._load(data)
        self.changed.connect(self.onChanged)
        self.changed.emit()

    @property
    def root(self):
        return self._root

    def _load(self, data):
        self._root = data if data else self._init()

    def prepare(self, root):
        pass

    def save(self):
        pass

    def _init(self):
        root = {}
        self.prepare(root)
        return root

    def onChanged(self):
        self.save()

    def __str__(self):
        return json.dumps(self.root)

    def serialize(self, fd):
        return json.dump(self.root, fd, indent=4)

    def dump(self):
        self.serialize(sys.stdout)

class QJSONFile(QJSONObj):
    def __init__(self, path, **kw):
        self._path = path

        super(QJSONFile, self).__init__(**kw)

    def _init(self):
        try:
            root = json.load(open(self.path, 'rt'))
        except Exception as e:
            root = collections.OrderedDict()
        self.prepare(root)

        return root

    @property
    def path(self):
        return self._path

    def save(self):
        with open(self.path, 'w+t') as fd:
            self.serialize(fd)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/ipfsinfos.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_IPFSInfosDialog(object):
    def setupUi(self, IPFSInfosDialog):
        IPFSInfosDialog.setObjectName("IPFSInfosDialog")
        IPFSInfosDialog.resize(428, 368)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(IPFSInfosDialog)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabWidget = QtWidgets.QTabWidget(IPFSInfosDialog)
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.tab)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout.setObjectName("gridLayout")
        self.repoSize = QtWidgets.QLabel(self.tab)
        self.repoSize.setText("")
        self.repoSize.setObjectName("repoSize")
        self.gridLayout.addWidget(self.repoSize, 1, 1, 1, 1)
        self.label_5 = QtWidgets.QLabel(self.tab)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 2, 0, 1, 1)
        self.label = QtWidgets.QLabel(self.tab)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.repoObjCount = QtWidgets.QLabel(self.tab)
        self.repoObjCount.setText("")
        self.repoObjCount.setObjectName("repoObjCount")
        self.gridLayout.addWidget(self.repoObjCount, 0, 1, 1, 1)
        self.repoMaxStorage = QtWidgets.QLabel(self.tab)
        self.repoMaxStorage.setText("")
        self.repoMaxStorage.setObjectName("repoMaxStorage")
        self.gridLayout.addWidget(self.repoMaxStorage, 2, 1, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.tab)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 1, 0, 1, 1)
        self.label_7 = QtWidgets.QLabel(self.tab)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 3, 0, 1, 1)
        self.repoVersion = QtWidgets.QLabel(self.tab)
        self.repoVersion.setText("")
        self.repoVersion.setObjectName("repoVersion")
        self.gridLayout.addWidget(self.repoVersion, 3, 1, 1, 1)
        self.gridLayout_2.addLayout(self.gridLayout, 0, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_2.addItem(spacerItem, 1, 0, 1, 1)
        self.tabWidget.addTab(self.tab, "")
        self.tab_2 = QtWidgets.QWidget()
        self.tab_2.setObjectName("tab_2")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.tab_2)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.gridLayout_3 = QtWidgets.QGridLayout()
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.label_6 = QtWidgets.QLabel(self.tab_2)
        self.label_6.setObjectName("label_6")
        self.gridLayout_3.addWidget(self.label_6, 1, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.tab_2)
        self.label_2.setObjectName("label_2")
        self.gridLayout_3.addWidget(self.label_2, 0, 0, 1, 1)
        self.agentVersion = QtWidgets.QLabel(self.tab_2)
        self.agentVersion.setText("")
        self.agentVersion.setObjectName("agentVersion")
        self.gridLayout_3.addWidget(self.agentVersion, 1, 1, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_3.addItem(spacerItem1, 3, 0, 1, 1)
        self.nodeId = QtWidgets.QLabel(self.tab_2)
        self.nodeId.setText("")
        self.nodeId.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.nodeId.setObjectName("nodeId")
        self.gridLayout_3.addWidget(self.nodeId, 0, 1, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.tab_2)
        self.label_4.setObjectName("label_4")
        self.gridLayout_3.addWidget(self.label_4, 2, 0, 1, 1)
        self.protocolVersion = QtWidgets.QLabel(self.tab_2)
        self.protocolVersion.setText("")
        self.protocolVersion.setObjectName("protocolVersion")
        self.gridLayout_3.addWidget(self.protocolVersion, 2, 1, 1, 1)
        self.verticalLayout_3.addLayout(self.gridLayout_3)
        self.tabWidget.addTab(self.tab_2, "")
        self.verticalLayout.addWidget(self.tabWidget)
        self.verticalLayout_2.addLayout(self.verticalLayout)
        self.okButton = QtWidgets.QPushButton(IPFSInfosDialog)
        self.okButton.setObjectName("okButton")
        self.verticalLayout_2.addWidget(self.okButton)

        self.retranslateUi(IPFSInfosDialog)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(IPFSInfosDialog)

    def retranslateUi(self, IPFSInfosDialog):
        _translate = QtCore.QCoreApplication.translate
        IPFSInfosDialog.setWindowTitle(_translate("IPFSInfosDialog", "Dialog"))
        self.label_5.setText(_translate("IPFSInfosDialog", "Maximum storage"))
        self.label.setText(_translate("IPFSInfosDialog", "Objects count"))
        self.label_3.setText(_translate("IPFSInfosDialog", "Repository size"))
        self.label_7.setText(_translate("IPFSInfosDialog", "Version"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _translate("IPFSInfosDialog", "Repository"))
        self.label_6.setText(_translate("IPFSInfosDialog", "Agent version"))
        self.label_2.setText(_translate("IPFSInfosDialog", "ID"))
        self.label_4.setText(_translate("IPFSInfosDialog", "Protocol version"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), _translate("IPFSInfosDialog", "Node"))
        self.okButton.setText(_translate("IPFSInfosDialog", "OK"))

